import inspect
from reprlib import recursive_repr
from datetime import datetime
from collections.abc import Mapping

from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field as dataclass_field, asdict

from spintop.utils import GetRecursiveMixin, GetRecursive, classproperty

from .view import DataClassPrimitiveView, ComplexPrimitiveView, join_key_prefix
from .meta import FieldsOf, MetaField
from .serialization import DataClassSchema, register_type, cls_of_obj, ValidationError, get_base_schema_cls, dump_json_schema

class CacheByCls():
    def __init__(self):
        self.value_by_cls = {}
        
    def retrieve(self, cls):
        if not self.value_by_cls.get(cls, None):
            self.value_by_cls[cls] = self._get(cls)
        return self.value_by_cls[cls]
    
    def _get(self, cls):
        raise NotImplementedError()
    
class FieldsCache(CacheByCls):
    """The fields() function is quite slow. This allows to store its 
    return value per cls, as it shouldn't change dynamically during
    runtime."""
    
    def _get(self, cls):
        return fields(cls)
    
class DefaultViewsCache(CacheByCls):
    def _get(self, cls):
        return DataClassPrimitiveView(
            cls.get_default_view_mapping()
        )

_fields_cache = FieldsCache()
_default_views_cache = DefaultViewsCache()

BaseDataClass = None # temporary define

def field_metadata(readonly=False, validate=None):
    return {
        'dump_only': readonly, # Marshmallow field key
        'validate': validate
    }

def is_field_readonly(field):
    return is_metadata_readonly(field.metadata)

def is_metadata_readonly(metadata):
    return metadata.get('dump_only', False)

def validate_field_value(field, value):
    validate = field.metadata.get('validate', None)
    if validate:
        validate(value)

def _model_dataclass(_register_type=None, schema_cls=None, **kwargs):
    """Custom dataclass decorator that removes the default values set on classes.
    This allows the meta base data class to return the field itself instead of the
    default value."""
    def _decorator(cls):
        
        # if not _base_cls and not issubclass(cls, BaseDataClass):
        #     cls = type(cls.__name__, (cls, BaseDataClass), {'__annotations__': get_annotations(cls)})

        model_properties = {}
        # Extract all properties from the data class and replace them by their
        # respective fields. 
        annotations = get_annotations(cls)

        for attr_name in dir(cls):
            if isinstance(getattr(cls, attr_name), _AbstractModelAttribute):
                return_type = annotations.get(attr_name, None)
                # If it has no return type, then it probably is not the fget accessor.
                if return_type:
                    prop = getattr(cls, attr_name)
                    model_properties[attr_name] = prop

                    prop.set_return_type(return_type)
                    setattr(cls, attr_name, prop.field)

        cls = dataclass(init=False, repr=False, **kwargs)(cls)

        # Once the dataclass is done and it detected the different fields,
        # replace the attributes by the final property accessor.
        for attr_name, prop in model_properties.items():
            value = prop.value_to_set_on_cls()
            if value is not None:
                setattr(cls, attr_name, value)

        register_type(cls, _register_type, schema_cls=schema_cls)
        return cls
    return _decorator

def get_annotations(obj):
    return obj.__dict__.get('__annotations__', {})

def set_annotations(obj, anno):
    obj.__dict__['__annotations__'] = anno

class _AbstractModelAttribute():
    def __init__(self, readonly=False, validators=[], **kwargs):
        self.readonly = readonly
        self.validators = validators
        self.field_kwargs = kwargs
        self._field = None
    
    @property
    def field(self):
        if self._field is None:
            kwargs = self.field_kwargs
            kwargs['metadata'] = self._metadata()
            self._field = dataclass_field(**kwargs)
        return self._field

    def set_return_type(self, return_type):
        pass

    def value_to_set_on_cls(self):
        return None

    def validate(self, value):
        for validator in self.validators:
            validator(value)
    
    def _metadata(self):
        return field_metadata(readonly=self.readonly, validate=self.validate)

class model_attribute(_AbstractModelAttribute):
    def __init__(self, default_factory, validators=[], **kwargs):
        super().__init__(default=default_factory, validators=validators, **kwargs)

class model_property(_AbstractModelAttribute):
    def __init__(self, fget=None, fset=None, fdel=None, doc=None, readonly=True, **kwargs):
        super().__init__(readonly=readonly, **kwargs)
        self.property = property(fget, fset, fdel, doc)

    def value_to_set_on_cls(self):
        return self.property

    def set_return_type(self, return_type):
        annotations = get_annotations(self.property.fget)
        annotations['return'] = return_type
        set_annotations(self.property.fget, annotations)

    def getter(self, fget):
        self.property = self.property.getter(fget)
        return self

    def setter(self, fset):
        self.property = self.property.setter(fset)
        return self

    def deleter(self, fdel):
        self.property = self.property.deleter(fdel)
        return self

def model_constant(value):
    return model_property(
        fget= lambda obj: value,
        readonly=True
    )
    
class MetaBaseDataClass(type):
    """Allows to replace a dataclass decorator with inheritance."""

    def __new__(cls, name, mro, attrs):
        _type = attrs.get('register_type_', attrs.get('__qualname__', name))
        # _base_cls = attrs.get('_base_cls', False)
        _schema_cls = attrs.get('schema_cls_', None)

        cls = super(MetaBaseDataClass, cls).__new__(cls, name, mro, attrs)
        return _model_dataclass(_type, schema_cls=_schema_cls)(cls)

class BaseDataClass(GetRecursiveMixin, metaclass=MetaBaseDataClass):
    register_type_ = '_base'
    schema_cls_ = None

    def __init__(self, *args, **kwargs):
        if args:
            raise NotImplementedError('Positionnal args not supported with BaseDataClass descendants.')
        defaults = self.get_defaults()

        self.update_attrs(defaults) # First set defaults
        self.update_attrs(kwargs) # Then constructor arguments.
        self.__post_init__()

        # Check if fields exist (or property)
        self.validate_attrs(kwargs)

    def __post_init__(self):
        pass

    @classmethod
    def defaults(cls, **others):
        return others

    @classmethod
    def get_defaults(cls, **other_fields):
        defaults = cls.defaults()
        defaults.update({
            field.name: 
                default_field_value(field) for field in cls.dataclass_fields() if field.name not in defaults and not is_field_readonly(field)
        })
        return defaults
    
    @classmethod
    def null(cls, **not_null_fields):
        """ Creates an instance with all attributes set to None. 
        Calls null() on any sub- BaseDataClass fields.
        """
        arguments = {}

        for key, default in cls.get_defaults().items():    
            arguments[key] = default
        
        arguments.update(not_null_fields)
        return cls(**arguments)
    
    @classmethod
    def dataclass_fields(cls):
        return _fields_cache.retrieve(cls)
    
    @classmethod
    def get_property_attrs(cls, with_setter_only=False):
        properties = set()
        for attr in dir(cls):
            value = getattr(cls, attr)
            if isinstance(value, property):
                if not with_setter_only or value.fset:
                    properties.add(attr)
        return properties

    @classmethod
    def data_dataclass_fields(cls):
        """ Sub classed by FeatureRecord below.
        """
        return cls.dataclass_fields()
        
    @classmethod
    def fields(cls):
        """Returns a FieldsOf object that allows to access the MetaFields of this cls.
        This allows a type based access to the underlying fields for query building.
        
        Examples:
            - Get the MetaField itself:
                >>> PersistenceRecord.fields().index.uuid
                MetaField(PersistenceRecord.index.uuid)

            - Get the complete dot separated name of this field:
                >>> PersistenceRecord.fields().index.uuid.name_
                'index.uuid'

            - Get the value of the MetaField of an instance of this type:
                >>> record = PersistenceRecord(index=PersistenceIDRecord(uuid='x'))
                >>> uuid_field = PersistenceRecord.fields().index.uuid
                >>> uuid_field.get_value(record)
                'x'
        """
        return FieldsOf(cls)

    @classmethod
    def get_default_view(cls):
        return _default_views_cache.retrieve(cls)
        
    @classmethod
    def get_default_view_mapping(cls, with_key_prefix=None):
        if with_key_prefix is not None:
            key_creator = lambda key: join_key_prefix(with_key_prefix, key)
        else:
            key_creator = lambda key: key
        
        # The default view uses the fields defined by the data_dataclass_fields class method.
        fields = cls.data_dataclass_fields()
        
        return {key_creator(field.name): MetaField(cls, field) for field in fields}

    @classmethod
    def get_schema(cls, schema_key=None, schema_type=None):
        if schema_type is None:
            schema_type = get_base_schema_cls(schema_key)
        return DataClassSchema(cls, schema_type)

    @classmethod
    def dump_json_schema(cls):
        return dump_json_schema(cls)

    @classmethod
    def from_dict(cls, data, schema_type=None):
        return cls.get_schema(schema_type)().load(data)

    def as_dict(self, schema_type=None):
        return self.get_schema(schema_type)().dump(self)

    def validate_attrs(self, other_attrs):
        # dataclass_fields = {field.name: field for field in self.dataclass_fields()}
        fields = {field.field_name_: field for field in self.fields()}

        unsupported_keys = set()

        other_attrs.update({key: getattr(self, key) for key in fields})

        for key, value in other_attrs.items():
            meta_field = fields.get(key)
            if not meta_field:
                possible_property = getattr(self.__class__, key, None)
                if isinstance(possible_property, property):
                    pass # OK.
                else:
                    # does not exist
                    unsupported_keys.add(key)
            else:
                field_obj = meta_field.field_
                if value is not None and isinstance(field_obj.type, type) and not isinstance(value, field_obj.type):
                    raise ValidationError(f'{self.__class__.__name__} attribute {key} has wrong type {type(value)}; requires {field_obj.type}')
                try:
                    validate_field_value(field_obj, value)
                except ValidationError as e:
                    raise ValidationError(f'{self.__class__.__name__} attribute {key} failed validation: {str(e)}')

        if unsupported_keys:
            settable_properties = self.get_property_attrs(with_setter_only=True)
            raise TypeError(f'{self.__class__.__name__} got unexpected argument(s) {unsupported_keys!r}. Fields are: {fields.keys()!r} and the following settable properties: {settable_properties!r}')

    def update_attrs(self, attributes_dict):
        for key, value in attributes_dict.items():
            setattr(self, key, value)
    
    def copy(self, with_data=True, as_cls=None, set_attributes={}):
        if with_data:
            nullify_fields = []
        else:
            nullify_fields = self.data_dataclass_fields()
            
        copy_obj = copy_nullify_fields(self, nullify_fields, target_cls=as_cls)
        for name, value in set_attributes.items():
            setattr(copy_obj, name, value)
        return copy_obj
    

    def getattr_from_field(self, meta_field):
        return getattr_from_field(self, meta_field)

    def hasattr_from_field(self, meta_field):
        return hasattr_from_field(self, meta_field)

    @recursive_repr()
    def __repr__(self):
        fields = tuple("{}={}".format(field.name, repr(getattr(self, field.name, None))) for field in self.dataclass_fields())
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))


def getattr_from_field(obj, meta_field):
    return meta_field.get_value(obj)

def hasattr_from_field(obj, meta_field):
    try:
        getattr_from_field(obj, meta_field)
        return True
    except AttributeError:
        return False


def copy_nullify_fields(obj, null_fields, target_cls=None):
    if target_cls is None:
        target_cls = obj.__class__
    
    copy_obj = target_cls.null()
    not_null_fields = (set(obj.dataclass_fields()) - set(null_fields)) & set(target_cls.dataclass_fields())
    for field in not_null_fields:
        setattr(copy_obj, field.name, getattr(obj, field.name))
    
    return copy_obj

def default_field_value(field):
    if not isinstance(field.default_factory, _MISSING_TYPE):
        value = field.default_factory()
    elif not isinstance(field.default, _MISSING_TYPE):
        if callable(field.default):
            value = field.default()
        else:
            value = field.default
    else:
        value = None
    
    return value

def default_cls_missing_view_fn(data_obj):
    cls = cls_of_obj(data_obj)
    return cls.get_default_view()

class DefaultPrimitiveView(ComplexPrimitiveView):
    def __init__(self, cls_missing_view_fn=None):
        if cls_missing_view_fn is None:
            cls_missing_view_fn = default_cls_missing_view_fn

        super(DefaultPrimitiveView, self).__init__(
            BaseDataClass, 
            cls_missing_view_fn=cls_missing_view_fn
        )


import datetime as dt
import traceback

from spintop.utils import utcnow_aware
from spintop.models import (
    BaseDataClass,
)

from spintop.logs import _logger

logger = _logger('filehandler')

class FileHandlingContext(object):
    def __init__(self, filename, env):
        self.env = env
        self.filename = filename
        self.start_event = None

    def new_event(self, event_type):
        return SpintopFileReference(
            filename=self.filename,
            pipeline_uuid=str(self.env.uuid),
            event=SpintopFileEvent(event_type=event_type)
        )

    def __enter__(self):
        self.start_event = self.new_event('start')
    
    def __exit__(self, exc_type, exc_value, traceback):
        end_event = self.new_event('end')
        logger.info('Finalizing file handling analytics.')
        if exc_type:
            logger.warning('Error occured, sending events then reraising...')
            end_event.event.severity = 'ERROR'
            end_event.event.data = {
                'exc_type': str(exc_type),
                'exc_value': str(exc_value),
                'traceback': str(traceback)
            }
        
        self.env.analytics_factory().processed_files.update([
            self.start_event,
            end_event
        ])

class SpintopFileEvent(BaseDataClass):
    start_datetime: dt.datetime = utcnow_aware
    event_type: str = 'start'
    data: dict = dict
    severity: str = 'INFO'
    count: int = 1

class SpintopFileReference(BaseDataClass):
    filename: str
    pipeline_uuid: str
    event: SpintopFileEvent = None



import os

from apache_beam.pipeline import StandardOptions
from apache_beam.options.pipeline_options import GoogleCloudOptions, SetupOptions 

SETUP_FILE = './setup.py'
REQUIREMENTS_FILE = './requirements.txt'

def create_dataflow_options(options, project_id, job_name, setup_file=None, requirements_file=None):
    # if not os.path.exists(SETUP_FILE):
    #     raise RuntimeError(f'Running on dataflow requires the {SETUP_FILE} file.')

    options = options.view_as(GoogleCloudOptions)
    options.project = project_id
    options.job_name = job_name
    options.temp_location = 'gs://nutaq-temp/dataflow-temp'
    options.staging_location = 'gs://nutaq-temp/dataflow-staging'
    options.view_as(StandardOptions).runner = 'DataflowRunner'
    
    if setup_file:
        if not isinstance(setup_file, str):
            # Can be bool
            setup_file = SETUP_FILE
        options.view_as(SetupOptions).setup_file = setup_file
    
    if requirements_file:
        if not isinstance(requirements_file, str):
            # Can be bool
            requirements_file = REQUIREMENTS_FILE
        options.view_as(SetupOptions).requirements_file = requirements_file



import requests

class DbtCloudInterface(object):
    def __init__(self, env):
        self.env = env
        self.token = env.DBT_CLOUD_AUTH_TOKEN
    
    def trigger_job(self, job_url, cause='Triggered from Spintop Data Analysis.', use_git_sha=True, use_git_branch=None):
        api_token = self.env.DBT_CLOUD_AUTH_TOKEN
        if not api_token:
            raise RuntimeError('DBT trigger requires DBT_CLOUD_AUTH_TOKEN in env.')

        body = {'cause': cause}
        if use_git_branch:
            body['git_branch'] = use_git_branch
        elif self.env.COMMIT_SHA and use_git_sha:
            body['git_sha'] = self.env.COMMIT_SHA

        r = requests.post(job_url, json=body, headers={'Authorization': f'Bearer {api_token}'})
        r.raise_for_status()


from json import dumps, loads

from marshmallow import fields

from google.cloud.bigquery import SchemaField

from spintop.models import AbstractSchema, register_schema_type
from spintop.models.serialization import DeepTabularCompatSchema

def bigquery_schema_from_model(model_cls):
    jsonschema = model_cls.dump_json_schema()
    return build_schema(jsonschema)
    
BigQuerySchema = DeepTabularCompatSchema
register_schema_type('bigquery', DeepTabularCompatSchema)

def define_schema(field, name):
    schema_name = name
    schema_type = "STRING"
    schema_mode = "NULLABLE"
    schema_description = None
    schema_fields = ()

    if 'type' not in field and 'anyOf' in field:
        for types in field['anyOf']:
            if types['type'] == 'null':
                schema_mode = 'NULLABLE'
            else:
                field = types
            
    if isinstance(field['type'], list):
        types = set(field['type'])
        if "null" in types:
            schema_mode = 'NULLABLE'
            types.remove("null")
        else:
            schema_mode = 'required'
        
        if len(types) != 1:
            raise ValueError('BigQuery only supports one type per property, with the exception of null.')

        schema_type = list(types)[0]
    else:
        schema_type = field['type']
    if schema_type == "object":
        schema_type, schema_fields = build_object_field(field)
    if schema_type == "array":
        schema_type = field.get('items').get('type')
        schema_mode = "REPEATED"
        if schema_type == "object":
          schema_type, schema_fields = build_object_field(field.get('items'))


    if schema_type == "string":
        if "format" in field:
            if field['format'] == "date-time":
                schema_type = "timestamp"

    if schema_type == 'number':
        schema_type = 'FLOAT'

    return (schema_name, schema_type, schema_mode, schema_description, schema_fields)

def build_object_field(field):
    if 'properties' in field:
        return ("RECORD", build_schema(field))
    else:
        # expect JSON
        return ("STRING", ())

def build_schema(schema):
    SCHEMA = []
    for key in schema['properties'].keys():
        
        if not (bool(schema['properties'][key])):
            # if we endup with an empty record.
            continue

        schema_name, schema_type, schema_mode, schema_description, schema_fields = define_schema(schema['properties'][key], key)
        SCHEMA.append(SchemaField(schema_name, schema_type, schema_mode, schema_description, schema_fields))

    return SCHEMA


class ForceTestbenchNameTransformer(object):
    def __init__(self, testbench_name):
        self.testbench_name = testbench_name
    
    def __call__(self, test_record):
        test_record.test_id.testbench.id = self.testbench_name
        return test_record


import os

from spintop_deploy.logs import _logger

logger = _logger('param-provider')
import os

NO_DEFAULT = object()

class BaseParamProvider(object):
    tolerate_defaults = True

    def get_environ_or_param_value(self, environ_name, param_name=None, decrypt=False, default=NO_DEFAULT):
        try:
            return self._get_environ_or_param_value(environ_name, param_name=param_name, decrypt=decrypt)
        except KeyError:
            if self.tolerate_defaults and default is not NO_DEFAULT:
                return default
            else:
                raise

    def _get_environ_or_param_value(self, environ_name, param_name=None, decrypt=False):
        value_from_environ = os.environ.get(environ_name)
        if value_from_environ is None:
            if param_name is None: param_name = environ_name
            return self.get_param_value(param_name, decrypt=decrypt)
        else:
            return value_from_environ

    def get_param_value(self, name, decrypt=False):
        raise NotImplementedError()

class EnvOnlyParamProvider(BaseParamProvider):
    def _get_environ_or_param_value(self, environ_name, param_name=None, decrypt=False):
        return os.environ[environ_name]

class SSMParamProvider(BaseParamProvider):
    """AWS Param provider"""
    
    def __init__(self, region=os.environ.get('AWS_REGION'), enable=True):
        import boto3
        self.ssm = boto3.client('ssm', region_name=region)
        self.enable = enable
    
    @classmethod
    def init(cls, config):
        return cls(
            region=config['AWS_REGION'],
            enable=not config['EMULATE_AWS']
        )
        
    def get_param(self, name, decrypt=False):
        if not self.enable:
            raise Exception('SSM is disabled when emulating AWS.')
        return self.ssm.get_parameter(
            Name=name,
            WithDecryption=decrypt
        )
    
    def get_param_value(self, name, decrypt=False):
        return self.get_param(name, decrypt=decrypt)['Parameter']['Value']
        

class GCPParamProvider(BaseParamProvider):
    
    def __init__(self, project_id, enable=True):
        from google.cloud import secretmanager
        self.client = secretmanager.SecretManagerServiceClient()
        self.enable = enable
        self.project_id = project_id

    def get_param_value(self, name, decrypt=False):
        return self.get_param(name)

    def get_param(self, name):
        if not self.enable:
            raise Exception('GCP Param Provider is disabled.')
        from google.api_core.exceptions import PermissionDenied
        try:
            path = self.client.secret_path(self.project_id, name) + '/versions/latest'
            return self.client.access_secret_version(name=path).payload.data.decode("utf-8")
        except PermissionDenied as e:
            raise KeyError(f'Unable to find secret with key {name}: {str(e)}')
        
    def new_param_version(self, name, value):
        parent_path = self.client.secret_path(self.project_id, name)
        value = value.encode('utf-8')
        self.client.add_secret_version(parent_path, {'data': value})


#!/usr/bin/env python3

import simplejson as json

from tempfile import TemporaryFile

from google.cloud import bigquery
from google.cloud.bigquery.job import SourceFormat
from google.cloud.bigquery import Dataset, WriteDisposition
from google.cloud.bigquery import LoadJobConfig
from google.api_core import exceptions

from spintop.logs import _logger

logger = _logger('bigquery')

class BigQueryDictLoadJobBuilder(object):
    def __init__(self, dataset_id, table_name, truncate=False, project_id=None, client=None, dataset=None):
        self.dataset_id = dataset_id
        self.table_name = table_name

        self.truncate = truncate
        self.schema = None
        self.data = TemporaryFile(mode='w+b')

        if client is None:
            client = bigquery.Client(project=project_id)

        if dataset is None:
            dataset = client.dataset(dataset_id)
        
        self.dataset = dataset
        self.client = client
    
    def add_row(self, row_dict):
        dat = bytes(json.dumps(row_dict) + '\n', 'UTF-8')
        self.data.write(dat)

    def add_rows(self, rows):
        for row in rows:
            self.add_row(row)
    
    @property
    def is_empty(self):
        return self.data.tell() == 0

    def build_job(self):
        dataset_ref = self.dataset
        table_ref = dataset_ref.table(self.table_name)

        load_config = LoadJobConfig()
        if self.schema is None:
            # No schema. Use Bigquery auto schema
            load_config.autodetect = True
        else:
            load_config.schema = self.schema

        load_config.source_format = SourceFormat.NEWLINE_DELIMITED_JSON
    
        if self.truncate:
            load_config.write_disposition = WriteDisposition.WRITE_TRUNCATE
        else:
            # Allow field addition or relaxation if not truncate.
            load_config.schema_update_options = [
                bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
                bigquery.SchemaUpdateOption.ALLOW_FIELD_RELAXATION
            ]
        
        file_ref = self.data
        data_bytes = file_ref.tell()
        file_ref.seek(0)
        load_job = self.client.load_table_from_file(
            file_ref, table_ref,
            job_config=load_config
        )
        logger.info(f'Created job {load_job.job_id} of {data_bytes/1000000} MBytes')
        return load_job


import os
import base64
import json
import tempfile

from shutil import copyfile
from binascii import Error as DecodeError
from collections.abc import Mapping

BYTES_ENCODING = 'utf-8'

def decode_data(data):
    raw_bytes = data
    # print(raw_bytes)
    endswith = b'=' if isinstance(data, bytes) else '='
    if data.endswith(endswith):
        try:
            raw_bytes = base64.b64decode(data)
        except DecodeError:
            pass # use raw_bytes = data
    
    try:
        string = raw_bytes.decode(BYTES_ENCODING)
    except (
            UnicodeDecodeError, # unable to decode
            AttributeError # 'str' object has no attribute 'decode'
        ):
        string = raw_bytes
    # print(string)
    return json.loads(string)

def encode_data(data):
    raw_bytes = json.dumps(data).encode(BYTES_ENCODING)
    return base64.b64encode(raw_bytes)

class Message(object):
    def __init__(self, event=None, context=None, data=None):
        if event is None:
            event = {'data': data}
        self._event = event
        self._context = context
        self._data = data

    @classmethod
    def from_message(cls, other_message):
        return cls(event=other_message.event, context=other_message.context)

    @property
    def event(self):
        return self._event

    @property
    def context(self):
        return self._context

    @property
    def attributes(self):
        if self.event:
            return self.event.get('attributes', {})
        else:
            return {}

    @property
    def event_type(self):
        return self.attributes.get('eventType')

    @property
    def data(self):
        if not self._data and self.event:
            # lazy load
            data = self.event.get('data')
            if isinstance(data, Mapping):
                # probably already decoded.
                self._data = data
            else:
                self._data = decode_data(data)

        return self._data

    def __repr__(self):
        return f'{self.__class__.__name__}(event={self.event},context={self.context})'

GCS_FINALIZE = 'OBJECT_FINALIZE'
GCS_DELETE = 'OBJECT_DELETE'

class GCSFileReferenceMessage(Message):

    @classmethod
    def from_message(cls, other_message):
        if other_message.event_type not in [GCS_FINALIZE, GCS_DELETE]:
            return None
        else:
            return super().from_message(other_message)

    @property
    def name(self):
        return self.attributes['objectId']

    @property
    def bucket(self):
        return self.attributes.get('bucketId', '__local__')

    def is_finalized(self):
        return self.event_type == GCS_FINALIZE

    def is_deleted(self):
        return self.event_type == GCS_DELETE

    def download_to_filename(self, filename):
        from google.cloud import storage
        storage_client = storage.Client()

        bucket = storage_client.bucket(self.bucket)
        blob = bucket.blob(self.name)
        blob.download_to_filename(filename)
    
    def download_temp(self):
        name = self.name
        if os.path.isabs(name):
            # Abs path. Use only filename for temp location or abspath will be used.
            _, name = os.path.split(name)
        
        filename = os.path.join(tempfile.gettempdir(), self.bucket, name)

        folder, _ = os.path.split(filename)
        os.makedirs(folder, exist_ok=True)
        
        self.download_to_filename(filename)
        return filename

class FileReferenceMessage(GCSFileReferenceMessage):
    def download_to_filename(self, filename):
        copyfile(self.name, filename)

class LogMessage(Message):

    @property
    def payload(self):
        return self.data.get('protoPayload', {})

    @property
    def service_name(self):
        return self.payload.get('serviceName')

    @property
    def service_data(self):
        return self.payload.get('serviceData')


class BigQueryAuditLogMessage(LogMessage):

    @classmethod
    def from_message(cls, other_message):
        message = super().from_message(other_message)
        if message.service_name == 'bigquery.googleapis.com':
            return message
        else:
            return None

    @property
    def bq_event_data(self):
        return self.service_data.get('jobCompletedEvent', {})

    @property
    def bq_job(self):
        return self.bq_event_data.get('job', {})

    @property
    def bq_job_config(self):
        return self.bq_job.get('jobConfiguration', {})

    @property
    def query_config(self):
        return self.bq_job_config.get('query', {})

    @property
    def job_status(self):
        return self.bq_job_config.get('jobStatus', {})

    @property
    def project_id(self):
        return self.query_config.get('destinationTable', {}).get('projectId')

    @property
    def dataset_id(self):
        return self.query_config.get('destinationTable', {}).get('datasetId')

    @property
    def table_id(self):
        return self.query_config.get('destinationTable', {}).get('tableId')

    @property
    def job_state(self):
        return self.job_status.get('state')

    def is_query_job_completed(self):
        return self.bq_event_data.get('eventName') == 'query_job_completed'



import pkgutil

version_file_content = pkgutil.get_data('spintop.compat', 'VERSION' )
VERSION = version_file_content.decode().strip()
    
def format_exception(e):
    return e.__class__.__name__ + ': ' + str(e)
    


from google.cloud import bigquery

from spintop.models import get_serializer
from spintop.services.bigquery.target_bigquery import emit_state, TargetBigQuery
from spintop.utils import repr_obj, utcnow_aware

from ..base import AbstractSingerTarget, SingerMessagesFactory, REQUEST_ID
from ..models import BigQueryAnalyticsResponse

class BigQuerySingerTarget(AbstractSingerTarget):
    add_null_to_fields = False

    def __init__(self, project_id, dataset_id, job_project_id=None, validate_records=False, truncate=False):
        super().__init__()
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.validate_records = validate_records
        self.truncate = truncate
        
        self.target_bigquery = TargetBigQuery.from_project_dataset_id(project_id=project_id, dataset_id=dataset_id, job_project_id=job_project_id)
        self.target_bigquery.try_create_dataset()

    def send_messages_dict(self, messages, request_id=None, stream_name=None):
        state = self.target_bigquery.persist_messages_job(messages, truncate=self.truncate, validate_records=self.validate_records)
        emit_state(state)
        return BigQueryAnalyticsResponse(
            count = len(messages),
            request_id = request_id,
            stream_name = stream_name,
            project_id = self.project_id,
            dataset_id = self.dataset_id
        )

    def __repr__(self):
        return repr_obj(self, ['project_id', 'dataset_id'])

class BigQueryAnalytics(BigQuerySingerTarget):
    def __init__(self, spintop_api=None, uri=None, database_name=None, env=None):
        try:
            job_project_id = env['GOOGLE_CLOUD_PROJECT']
        except KeyError:
            job_project_id = None
            
        super().__init__(project_id=uri, dataset_id=database_name, job_project_id=job_project_id)
        self._client = None

def query_analytics_response(resp, deserialize=True, client=None):
    if client is None:
        client = bigquery.Client()

    query = f"""
    SELECT * FROM `{resp.project_id}.{resp.dataset_id}.{resp.stream_name}` WHERE {REQUEST_ID}='{resp.request_id}'
    """
    query_job = client.query(query)  # Make an API request.
    
    records = list(dict(record) for record in query_job)

    if deserialize:
        records = [get_serializer('tabular').deserialize(r, partial=True) for r in records]

    return records



from enum import Enum
from functools import wraps
from collections import defaultdict
from collections.abc import Mapping

from collections import namedtuple

from .models import PersistenceRecordsUpdateMessage

Topic = namedtuple('Topic', ('name', 'message_cls'))

class Topics(Enum):
    RECORDS_UPDATE_TOPIC = Topic(name='records.update', message_cls=PersistenceRecordsUpdateMessage)

def messages_endpoint(topic):
    return property(
        fget = lambda publisher: _MessagesEndpoint(publisher, topic)
    )

class SpintopMessagePublisher(object):
    records_update = messages_endpoint(Topics.RECORDS_UPDATE_TOPIC)

    def subscribe(self, topic, callback):
        raise NotImplementedError()

    def publish(self, topic, message):
        raise NotImplementedError()

    def prepare_message(self, message):
        pass

    def supports_publish(self):
        return True

    def supports_subscribe(self):
        return True

class _MessagesEndpoint(object):
    def __init__(self, publisher, topic_obj):
        topic_obj = topic_obj.value
        
        topic = topic_obj.name
        message_model_cls = topic_obj.message_cls

        self.publisher = publisher
        self.topic = topic
        self.schema = message_model_cls.get_schema('json')()

    def subscribe(self, callback):
        if self.publisher.supports_subscribe():
            @wraps(callback)
            def _wrapper(raw_message):
                message = self.load_message(raw_message)
                callback(message)

            self.publisher.subscribe(self.topic, _wrapper)

    def dump_message(self, message):
        if isinstance(message, Mapping):
            return message
        else:
            return self.schema.dump(message)

    def load_message(self, raw_message):
        return self.schema.load(raw_message)

    def publish(self, message):
        if self.publisher.supports_publish():
            self.publisher.prepare_message(message)
            content = self.dump_message(message)
            self.publisher.publish(self.topic, content)

class LocalMessagePublisher(SpintopMessagePublisher):
    def __init__(self, local_env):
        self._local_subscribers = defaultdict(list)
        self._env = local_env

    def subscribe(self, topic, callback):
        self._local_subscribers[topic].append(callback)

    def publish(self, topic, message):
        self._publish(topic, message)
        self._notify_local(topic, message)

    def prepare_message(self, message):
        message.env = self._env.freeze_database_access_only()
        super().prepare_message(message)

    def _notify_local(self, topic, message):
        for subscriber in self._local_subscribers[topic]:
            subscriber(message)

    def _publish(self, topic, message):
        pass

class NoopMessagePublisher(SpintopMessagePublisher):
    
    def supports_publish(self):
        return False

    def supports_subscribe(self):
        return False


import os
import itertools
from functools import wraps

from spintop.errors import SpintopException
from spintop.utils import AnonymousGetRecursive


from pymongo import UpdateOne, MongoClient, ASCENDING, DESCENDING
from pymongo.errors import BulkWriteError

from .queryset import MongoQuery

def db_from_mongo_uri(uri, database_name):
    client = MongoClient(uri)
    return client[database_name]

class DuplicateKeyError(SpintopException):
    def __init__(self, unique_fields):
        super().__init__(f'Feature {unique_fields} is not unique.')

def parse_write_error(raw_details):
    if raw_details.get('code') == 11000:
        return DuplicateKeyError(raw_details.get('keyValue'))
    else:
        return SpintopException(str(raw_details))

class MongoOperations(object):
    def __init__(self, mongo_collection, grouped_by=None):
        self.ops = mongo_collection
        self.grouped_by = grouped_by
        
    def count(self, query, grouped_by=None):
        grouped_by = self._grouped_by_or_default(grouped_by)
        query_dict = MongoQuery(query).build()
        distinct_values = set()
        for doc in self.ops.find(query_dict, projection=[grouped_by], batch_size=5000):
            gettable_doc = AnonymousGetRecursive(doc)
            distinct_values.add(gettable_doc.get_recursive(grouped_by))
        return len(distinct_values)

    def count_total(self, query):
        return self.count(query, grouped_by='_id')

    def find(self, query, limit_range=None):
        query_dict = MongoQuery(query).build()
        pipeline = self._grouped_by_test_uuid_pipeline(query_dict)

        if limit_range:
            pipeline += [
                {"$skip": limit_range[0]},
                {"$limit": limit_range[1] - limit_range[0]}
            ]

        all_grouped = self.ops.aggregate(pipeline, allowDiskUse=True)
        return itertools.chain.from_iterable(grouped.get('documents', []) for grouped in all_grouped)

    def _grouped_by_or_default(self, grouped_by):
        if grouped_by is None:
            grouped_by = self.grouped_by if self.grouped_by else '_id'
        return grouped_by

    def _grouped_by_test_uuid_pipeline(self, query_dict, grouped_by=None):
        grouped_by = self._grouped_by_or_default(grouped_by)
        return [
            { "$match": query_dict },
            { "$group": {
                "_id": f"${grouped_by}",
                "documents": { "$push":  "$$ROOT" }
            }},
            { "$sort": {'_id': 1}}
        ]
        
    def insert_many(self, objs):
        try:
            return self.ops.insert_many(objs)
        except BulkWriteError as all_errors:
            errors = [parse_write_error(e) for e in all_errors.details.get('writeErrors', [])]

            if len(errors) == 1:
                raise errors[0]
            elif len(errors) > 1:
                messages = ['Multiple write errors occured'] + [str(e) for e in errors]
                raise SpintopException('\n'.join(messages))
            else:
                # Weird...
                raise
            
    
    def update_many(self, objs, obj_query=lambda obj: {'_id': obj['_id']}):
        updates = [
            UpdateOne(obj_query(obj), {'$set': obj, '$inc': {'version': 1}}, upsert=True) for obj in objs
        ]
        return self.ops.bulk_write(updates)
    
    def delete_many(self, query):
        query_dict = MongoQuery(query).build()
        return self.ops.delete_many(query_dict)
        
    def create_index(self, fields_and_direction, unique=False):
        self.ops.create_index(fields_and_direction, unique=unique)

    def collection_stats(self):
        return self.ops.database.command('collstats', self.ops.name)


from ._collection import (
    SpintopTestRecordCollection, 
    SpintopTestRecordBuilder,
    SpintopSerializedTestRecord,
    SpintopTestRecord,
    SpintopTestRecordView,
    SpintopSerializedTestRecordCollection,
    TestDataRecord
)

from ._base import (
    BaseDataClass,
    TestIDRecord,
    FeatureRecord,
    MeasureFeatureRecord,
    PhaseFeatureRecord,
    DutIDRecord,
    TestbenchIDRecord,
    TestRecordSummary,
    DutOp,
    OutcomeData
)

from ._tree_struct import SpintopTreeTestRecord, SpintopTreeTestRecordBuilder


from functools import lru_cache

from spintop.analytics.bigquery import BigQueryAnalytics
from .analytics_functions import create_dataset_functions

@lru_cache(100)
def _cached_create_dataset_functions(analytics):
    return create_dataset_functions(analytics)

class AdvancedBigQueryAnalytics(BigQueryAnalytics):

    @property
    def functions(self):
        # Cache so that self.functions always return the same object.
        return _cached_create_dataset_functions(self)


from copy import copy
from collections.abc import Mapping

from anytree import NodeMixin, RenderTree, PreOrderIter, LevelOrderGroupIter
from anytree.render import ContStyle

from ._collection import SpintopTestRecord, SpintopTestRecordBuilder
from ._base import TestIDRecord, OutcomeData, PhaseFeatureRecord, MeasureFeatureRecord, NAME_SEPARATOR
    
class SpintopTestRecordError(Exception): pass
    
class AbstractSpintopTestRecordNode(NodeMixin):
    internal_feature = None
    separator = NAME_SEPARATOR
    
    def __init__(self, name, parent=None, **data):
        self.data = {}
        
        self.name = name
        self.parent = parent
        if data:
            self.define_data(**data)
        
    @property
    def tree_name(self):
        return self.separator.join(p.name for p in self.path)

    def __getattr__(self, name):
        try:
            return self.data[name]
        except KeyError as e:
            raise AttributeError(str(e))

    def child_tree_name(self, child_name):
        return self.separator.join([self.tree_name, child_name])
    
    def define_id(self,
            **kwargs
        ):

        self.root.define_data(
            **kwargs
        )
        
        
    def define_data(self, **data):
        for key in data:
            # Will raise AttributeError if it does not exist.
            self.internal_feature.fields().get_field(key)

        self._update_data(**data)
        
    def _update_data(self, **data):
        self.data.update(data)
        
    def __repr__(self):
        return "%s(name='%s')" % (self.__class__.__name__, self.name)
    
    def __copy__(self):
        return self.__class__(self.name, parent=None, **self.data)
    
    def to_internal(self, index=0):
        internal_cls = self.internal_feature
        obj = internal_cls(
            oid=None, # New class, not bound to any stored one.
            name=self.name,
            version=0,
            depth=self.depth,
            index=index,
            ancestors=[node.name for node in self.path[:-1]],
            original=True,
            user_data={},
            **self._data_for_internal()
        )
        return obj
        
    def _data_for_internal(self):
        return self.data
        
class Phase(AbstractSpintopTestRecordNode):
    internal_feature = PhaseFeatureRecord
    
class Measure(AbstractSpintopTestRecordNode):
    internal_feature = MeasureFeatureRecord
        
class SpintopTreeTestRecord(Phase):
    internal_feature = TestIDRecord

    def __init__(self):
        super(SpintopTreeTestRecord, self).__init__('', parent=None)

    def add_tag(self, key, value=True):
        if not self.data.get('tags'):
            self.define_data(tags={})

        self.data['tags'][key] = value

    def _data_for_internal(self):
        return dict(
            **self.data
        )
    
    def flatten_data(self):
        builder = SpintopTestRecordBuilder(
            summary = self.to_internal(index=0),
            features = [node.to_internal(index) for index, node in enumerate(PreOrderIter(self)) if node is not self]
        )
        return builder.build()
    
    def pprint(self):
        print(RenderTree(self, style=ContStyle()))
        
    def group_outcomes(self, depth=1):
        for nodes in LevelOrderGroupIter(self, maxlevel=depth):
            if nodes[0].depth < depth-1:
                continue
            else:
                current_group = []
                for node in nodes:
                    if node.data.get('outcome', True):
                        current_group.append(node)
                    else:
                        if current_group:
                            yield current_group, True
                            current_group = []
                    
                        yield [node], False
        
                if current_group:
                    yield current_group, True
                
        

class SpintopTestRecordBuilderContext(object):
    def __init__(self, ensure_unique_names=True):
        self._should_ensure_unique_names = ensure_unique_names
        self._duplicate_key_index = {}
        self._unique_ensured_nodes = {}
        
    def ensure_unique_key(self, key, parent):
        unique_name = parent.child_tree_name(key)
        if self._should_ensure_unique_names and unique_name in self._unique_ensured_nodes:
            suffix = self._duplicate_key_index.get(unique_name, 0)
            self._duplicate_key_index[unique_name] = suffix + 1
            key = key + str(suffix)
            return self.ensure_unique_key(key, parent)
        return key
    
    def add_unique_node(self, node):
        self._unique_ensured_nodes[node.tree_name] = node
    
class SpintopTreeTestRecordBuilder(object):
    def __init__(self, node=None, context=None, **context_kwargs):
        if node is None:
            node = SpintopTreeTestRecord()
        if context is None:
            context = SpintopTestRecordBuilderContext(**context_kwargs)
            
        self.node = node
        self.context = context
    
    def __enter__(self):
        if self.node.parent is None:
            raise SpintopTestRecordError('Cannot use the top level node as a context manager')
        return self
    
    def __exit__(self, *args, **kwargs):
        pass
    
    def __getattr__(self, name):
        return getattr(self.node, name)

    @property
    def test_record(self):
        return self.node.root

    @property
    def parent(self):
        return self._wrap(self.node.parent)
    
    def build(self):
        return self.test_record.flatten_data()
    
    def set_top_level_information(self, *args, **kwargs):
        self.define_id(*args, **kwargs)
        self.define_data(uuid=self.to_internal().uuid)
        
    def new_phase(self, name, outcome, duration):
        return self._new_node(Phase, name,
            outcome=outcome,
            duration=duration
        )
    
    def new_measure(self, name, outcome, value):
        return self._new_node(Measure, name,
            outcome=outcome,
            value=value
        )
    
    def _new_node(self, cls, name, **data):
        name = self.context.ensure_unique_key(name, parent=self.node)
        node = cls(name, parent=self.node, **data)
        self.context.add_unique_node(node)
        return self._wrap(node)
    
    def _wrap(self, node):
        return self.__class__(node, self.context)
    



from .base import record_to_lifecycle_step, feature_to_lifecycle_feature, reserve_lifecycle_streams
from .features import LCFeature
from .files import FileHandlingContext
from .steps import LCStep, LCUnit, LCTypes


""" Base models that allow persistency and indexing using the spintop.persistence modules.
"""
from datetime import datetime
from collections import OrderedDict
from typing import List, Dict, Optional, Any, NamedTuple

from .base import BaseDataClass, model_property, model_attribute, DefaultPrimitiveView
from .common import OutcomeData, DutIDRecord, VersionIDRecord
from .serialization import SerializedWrapper, TYPE_KEY, ValidationError, cls_subclasses, cls_of_serialized
from .queries import BaseQuery
from .meta import FieldsOf
from .view import DataClassPrimitiveView

from spintop.utils import utcnow_aware, is_aware, dict_ops, is_valid_py_identifier, isnan

def validate_tags_are_py_identifers(tags):
    for key in tags:
        if not is_valid_py_identifier(key):
            raise ValidationError(f'{key!r} is not a valid python identifier.')

def create_uuid():
    return uuid4()

def uuid_to_slug(uuidstring):
    try:
        uuid = UUID(uuidstring)
    except AttributeError:
        uuid = uuidstring
    
    return urlsafe_b64encode(uuid.bytes).rstrip(b'=').decode('ascii')

def uuid_generator(*parts):
    parts_str = []

    for part in parts:
        if hasattr(part, 'strftime'):
            # datetime
            parts_str.append(part.strftime("%Y%m%d-%H%M%S-%f"))
        else:
            parts_str.append(str(part))

    if not parts_str:
        # Not enough info for unique id. Generate one.
        return uuid_to_slug(create_uuid())
    else:
        return '.'.join(parts_str)

class PersistenceIDRecord(BaseDataClass):
    uuid: str
    is_deleted: bool = False
    tags: Dict[str, Optional[Any]] = model_attribute(dict, validators=[validate_tags_are_py_identifers])
    start_datetime: datetime = utcnow_aware

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.start_datetime and not is_aware(self.start_datetime):
            raise ValueError('Please provide a tz-aware datetime.')


class PersistenceRecord(BaseDataClass):
    # Common to all
    uuid: str = model_property(fget=lambda record: record.index.uuid)

    # Same attributes, type should be changed by sub classes.
    index: PersistenceIDRecord = None
    data: BaseDataClass = None

class SerializedPersistenceRecord(SerializedWrapper):
    pass

class ProcessStepIDRecord(PersistenceIDRecord):
    step: VersionIDRecord = None
    station: VersionIDRecord = None
    dut: DutIDRecord = DutIDRecord
    operator: str = None
    stage: str = 'prod'
    duration: float = None
    outcome: OutcomeData = None
    
    def __post_init__(self):
        super().__post_init__()
        self.dut = DutIDRecord.create(self.dut)
        
        if self.step is not None:
            self.step = VersionIDRecord.create(self.step)
        
        if self.station is not None:
            self.station = VersionIDRecord.create(self.station)

        if self.outcome is not None:
            self.outcome = OutcomeData.create(self.outcome)
            
        for key, value in self.tags.items():
            self.add_tag(key, value)

        if self.uuid is None:
            self.uuid = uuid_generator(self.step, self.start_datetime, self.dut)
            

    def add_tag(self, key, value=True):
        self.tags[key] = tag_value_sanitizer(value)
        
    def remove_tag(self, key):
        del self.tags[key]
        
def tag_value_sanitizer(value):
    if isinstance(value, float) and isnan(value):
        # NaN to None.
        value = None

    return value

class ProcessStepRecord(PersistenceRecord):
    index: ProcessStepIDRecord = None

class QualityStepIDRecord(ProcessStepIDRecord):
    """Outcome now included in ProcessStepIDRecord"""

class QualityStepRecord(ProcessStepRecord):
    """Outcome now included in ProcessStepIDRecord"""

class AssemblyStepIDRecord(PersistenceIDRecord):
    duts: List[DutIDRecord] = list

    def __post_init__(self):
        super().__post_init__()
        self.duts = [DutIDRecord.create(dut) for dut in self.duts]

class AssemblyStepRecord(PersistenceRecord):
    index: AssemblyStepIDRecord = None

index_fields = PersistenceIDRecord.fields()

def all_query_fields_dict():
    all_fields = list(PersistenceIDRecord.fields())

    for subcls in cls_subclasses(PersistenceIDRecord):
        all_fields += list(subcls.fields())

    return {field.field_name_: field for field in all_fields}

class PersistenceRecordCollection(BaseDataClass):
    records: List[PersistenceRecord] = list

    def __init__(self, records=None):
        records = list(records) if records else []
        super().__init__(records=sorted(records, key=lambda record: index_fields.uuid.get_value(record.index)))
    
    @property
    def collector(self):
        return self.add_record

    def add_record(self, record):
        self.records.append(record)
        
    def apply(self, *fns):
        for record in self.iter_records():
            for fn in fns:
                fn(record)
                
    def iter_records(self):
        for record in self.records:
            yield record
                
    def count_unique(self, key=lambda record: record.uuid):
        occurances = set()
        self.apply(lambda record: occurances.add(key(record)))
        return len(occurances)
            
    def sort(self, key=lambda record: record.index.start_datetime):
        self.records.sort(key=key)
        
    def __eq__(self, other):
        return self.records == other.records

class Query(BaseQuery):
    
    def default_parts(self):
        yield index_fields.is_deleted == False

    @classmethod
    def from_q_string(cls, q_string):
        return Query(eval(q_string, {}, all_query_fields_dict()))


class PersistenceRecordView(object):
    data_prefix = ('data',)
    index_prefix = ()
    default_view = DefaultPrimitiveView()
    
    def __init__(self, data_view=None, include_index=True, include_data=True, data_prefix=None, index_prefix=None):
        self.include_index = include_index
        self.include_data = include_data

        if data_view:
            self.data_view = DataClassPrimitiveView(data_view)
        else:
            self.data_view = None

        if data_prefix is not None:
            self.data_prefix = data_prefix
        
        if index_prefix is not None:
            self.index_prefix = index_prefix

    def apply(self, record, flatten_dict=True):
        result = OrderedDict()
        
        if self.include_index:
            index_data = self.apply_index(record, flatten_dict=flatten_dict, key_prefix=self.index_prefix)
            dict_ops.update(result, index_data)
        
        if self.data_view:
            data = self.apply_data(record, flatten_dict=flatten_dict, key_prefix=self.data_prefix)
            dict_ops.update(result, data)
        
        result = dict_ops.replace_defaults(result)

        return result
    
    def apply_index(self, record, **apply_kwargs):
        return self.default_view.apply(record.index, **apply_kwargs)

    def apply_data(self, record, **apply_kwargs):
        return self.data_view.apply(record.data, **apply_kwargs)

def cls_of_record(record):
    if isinstance(record, BaseDataClass):
        return record.__class__

    if isinstance(record, SerializedPersistenceRecord):
        data = record.as_dict()
    else:
        data = record

    if data:
        return cls_of_serialized(data)
    else:
        return None


import os
import itertools
import datetime as dt
import hashlib

from google.cloud import bigquery

from jinja2 import Environment, FileSystemLoader, PrefixLoader

from spintop.models import Filter, QueryResult
from spintop.logs import _logger

from ..base import get_uuid_datetime

HERE = os.path.abspath(os.path.dirname(__file__))

logger = _logger('bq-analytics')

def create_dataset_functions(analytics_target):
    bigquery_client = analytics_target.target_bigquery.bigquery_client
    dataset = analytics_target.target_bigquery.dataset

    return BigQueryDatasetFunctions(bigquery_client, dataset)

DEFAULT_SELECT_FIELDS = [
    '* EXCEPT (_type)'
]

class BigQueryDatasetFunctions(object):
    def __init__(self, client, dataset):
        self.client = client
        self.dataset = dataset

        self._existing_tables = None

        self._loaders = {
            'core': FileSystemLoader(os.path.join(HERE, 'queries'))
        }

        self.queries = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=PrefixLoader(self._loaders)
        )

    @property
    def datatransfer_parent_path(self):
        return self.datatransfer_client.project_path(self.client.project) 

    def add_template_loader(self, key, loader):
        self._loaders[key] = loader

    def map_stream_to_latest_views(self, stream_table, extract_tags=[]):
        return self.render_template('core/stream_to_last.bql', 
            from_table=self.fully_qual_table_name(stream_table),
            extract_tags=extract_tags
        )

    def helper_functions(self):
        return self.render_template('core/create_helper_functions.bql')

    def lc_unit_yields(self, origin_table, lc_type, partition_by=(), select_fields=None):
        partition_by = ('unit.id',) + tuple(partition_by)
        if select_fields is None:
            select_fields = DEFAULT_SELECT_FIELDS
            
        return self.render_template('core/lc_unit_yields.bql',
            from_table=self.fully_qual_table_name(origin_table),
            lc_type=lc_type,
            partition_by=partition_by,
            select_fields=select_fields
        )

    def lc_multi_transitions(self, origin_table, group_by_lc_types=('step',)):
        return self.render_template('core/multi_transitions.bql',
            from_table=self.fully_qual_table_name(origin_table),
            group_by_lc_types=group_by_lc_types
        )

    def stored_query(self, project_id, dataset, query_name, params={}, queries_table='queries'):
        return self.render_template('core/execute_stored_query.bql',
            project_id=project_id,
            dataset=dataset,
            queries_table=queries_table,
            query_name=query_name,
            params=params
        )

    def retrieve_one_uuid_partition(self, project_id, dataset, from_table, uuid, columns=None, input_safe=False):
        uuid_datetime = get_uuid_datetime(uuid)
        DATE_FORMAT = '%Y-%m-%d'
        
        table_schema = None

        if columns is None:
            # uuid is passed as a query param, so no chance of sql injection.
            input_safe = True

        if not input_safe:
            table_name = fully_qual_table_name(project_id, dataset, from_table)
            table = self.client.get_table(table_name)
            table_schema = table.schema

        if columns:
            columns = safe_columns(columns, table_schema=table_schema)

        return self.render_template('core/retrieve_one_uuid_partition.bql',
            from_table=fully_qual_table_name(project_id, dataset, from_table),
            query_params={'uuid': ('STRING', uuid)},
            lower_date_iso=uuid_datetime.strftime(DATE_FORMAT),
            upper_date_iso=(uuid_datetime + dt.timedelta(days=1)).strftime(DATE_FORMAT),
            columns=columns
        )

    def select_many_filtered_ordered(self, project_id, dataset, from_table, filters=[], order_by=[], columns=None, limit_rows=None, input_safe=False):
        table_schema = None

        if not input_safe:
            table_name = fully_qual_table_name(project_id, dataset, from_table)
            table = self.client.get_table(table_name)
            table_schema = table.schema

        if columns:
            columns = safe_columns(columns, table_schema=table_schema)

        if order_by:
            order_by = spintop_order_by_to_bq(order_by, table_schema=table_schema)

        bq_filters, query_params = spintop_filters_to_bq(filters, table_schema=table_schema)

        return self.render_template('core/select_many_filtered_ordered.bql',
            from_table=table_name,
            query_params=query_params,
            filters=bq_filters,
            order_by=order_by,
            columns=columns,
            limit_rows=limit_rows
        )

    def measure_values_mart(self, origin_table, measure_names={}, column=None):
        if column is None:
            column = 'value_f'
        if measure_names:
            return self.render_template('core/measures_mart.bql', 
                from_table=self.fully_qual_table_name(origin_table), 
                measure_names=measure_names,
                column=column
            )
        else:
            logger.info('Skipped measure mart because measure names are empty.')

    def array_measure_declarative_values_mart(self, origin_table, declarative_measures={}):
        """
        Declarative measures are of style
        
        # yaml
        ---
        Harmonics:
            keys: 
                freq: [70, 200, 1000, 2000, 3000]
                tx: [1, 2, 3, 4]
            name_format: 'Tx{tx}_{freq}MHz_Harmonic2'
        
        """
        return self.array_measure_values_mart(origin_table,
            measures=array_measure_generator_from_declarative(declarative_measures)
        )

    def array_measure_values_mart(self, origin_table, measures={}):
        """
        Normal array measures are of style

        # yaml
        ---
        Harmonics:
          - name: 'Tx{tx}_{freq}MHz_Harmonic2'
            keys:
              - freq: 70
                tx: 1
              - freq: 70
                tx: 2
              - ...
            value_column: value_f
        """
        if measures:
            return self.render_template('core/array_measures_mart.bql', 
                from_table=self.fully_qual_table_name(origin_table), 
                measures=measures
            )
        else:
            logger.info('Skipped array measure mart because measures are empty.')

    def list_tables(self):
        if not self._existing_tables:
            self._existing_tables = {table.table_id: table for table in self.client.list_tables(self.dataset)}
        return self._existing_tables

    def table_exists(self, table_name):
        dataset_tables = self.list_tables() 
        table_ref = dataset_tables.get(table_name, None)
        return table_ref is not None

    def create_empty_table(self, table_name):
        table_name = self.fully_qual_table_name(table_name)
        table = bigquery.Table(table_name)
        table = self.client.create_table(table)  # Make an API request.

    def render_template(self, query_template, query_params={}, **template_kwargs):
        template = self.queries.get_template(query_template)
        query = template.render(repr=repr, fully_qual_table_name=self.fully_qual_table_name, **template_kwargs)
        return self.prepare_query(query, name=query_template, query_params=query_params)

    def prepare_query(self, query, name=None, query_params={}):
        return PreparedQuery(query, self, name=name, params=query_params)

    def update_view(self, table_name, prepared_query):
        logger.info(f'Updating view {prepared_query} to {table_name!r}')

        query = prepared_query.query
        dataset_tables = self.list_tables() 
        table_ref = dataset_tables.get(table_name, None)

        exists = False
        update = False

        if table_ref:
            # Exists
            exists = True
            view = self.client.get_table(table_ref)

            if view.view_query != query:
                logger.info(f'View table {table_name!r} exists but query does not match; will update.')
                update = True
        else:
            # Does not exists.
            logger.info(f'View table {table_name!r} does not exist; will create.')
            exist = False
            update = True
            view_ref = self.dataset.table(table_name)
            view = bigquery.Table(view_ref)

        if update:
            view.view_query = query
            try:
                if exists:
                    self.client.update_table(view, ["view_query"])
                else:
                    self.client.create_table(view)
            except:
                logger.error(f'View with query \n {query} \n failed.', exc_info=True)
                raise

    def materialize_query(self, table_name, prepared_query):
        logger.info(f'Materializing {prepared_query} to {table_name!r}')

        query = prepared_query.query
        
        dataset_tables = self.list_tables() 
        table_ref = dataset_tables.get(table_name, None)

        if table_ref:
            logger.info(f'Table exists, dropping.')

        job_config = bigquery.QueryJobConfig(destination=self.fully_qual_table_name(table_name))
        job_config.write_disposition = "WRITE_TRUNCATE"
        # Start the query, passing in the extra configuration.
        logger.info(f'View will be updated; will not wait for query to end.')
        query_job = self.client.query(query, job_config=job_config)  # Make an API request.
        # query_job.result()  # Wait for the job to complete.
        return query_job

    def execute_query(self, prepared_query):
        logger.info(f'Executing {prepared_query}')

        query = prepared_query.query
        job_config = bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ScalarQueryParameter(param_name, param_type, value)
                for param_name, (param_type, value) in prepared_query.params.items()
            ]
        )

        query_job = self.client.query(query, job_config=job_config)  # Make an API request.
        return BigQueryResult(query_job, self.client)

    def fully_qual_table_name(self, table_name):
        return fully_qual_table_name(self.dataset.project, self.dataset.dataset_id, table_name)

    def create_gsheets_table(self, table_name, sheet_uri, schema, skip_leading_rows=0, sheet_range=None):
        logger.info(f'Creating external sheets source table {table_name}')

        dataset_tables = self.list_tables() 
        table_ref = dataset_tables.get(table_name, None)

        if table_ref:
            logger.info(f'Table exists, dropping.')
            self.client.delete_table(table_ref)  # Make an API request.

        table = bigquery.Table(self.fully_qual_table_name(table_name), schema=schema)
        external_config = bigquery.ExternalConfig("GOOGLE_SHEETS")
        external_config.source_uris = [sheet_uri]
        external_config.options.skip_leading_rows = skip_leading_rows
        external_config.options.range = sheet_range
        table.external_data_configuration = external_config

        return self.client.create_table(table)

def fully_qual_table_name(project_id, dataset_id, table_name):
    return '.'.join([project_id, dataset_id, table_name])

class PreparedQuery(object):
    def __init__(self, query, functions, name=None, params={}):
        self.query = query
        self.functions = functions
        self.name = name
        self.params = params

    def update_view(self, table_name):
        return self.functions.update_view(table_name, self)
    
    def materialize_query(self, table_name):
        return self.functions.materialize_query(table_name, self)

    def execute_query(self):
        return self.functions.execute_query(self)

    def __str__(self):
        name = self.name

        if not name:
            name = (self.query[:75] + '..') if len(self.query) > 77 else self.query
        
        return f'{self.__class__.__name__}({name})'

class BigQueryResult(object):
    def __init__(self, query_job, client):
        self.client = client
        self.query_job = query_job

    def __iter__(self):
        return iter(self.query_job)

    # def list_rows(self, max_results=None):
    #     destination_id = self.query_job.destination

    #     self.query_job.result() # Wait for it to end before retrieving results.
    #     destination = self.client.get_table(destination_id)
    #     iterator = 
    #     for row in self.client.list_rows(destination, max_results=max_results):
    #         yield dict(row)

    def as_query_result(self, max_results=None):
        destination_id = self.query_job.destination

        self.query_job.result() # Wait for it to end before retrieving results.

        # Is this necessary ? API call
        # destination = self.client.get_table(destination_id)
        destination = self.query_job.destination

        row_iterator = self.client.list_rows(destination, max_results=max_results)
        rows = [dict(row) for row in row_iterator]
        return QueryResult(
            query_id=self.query_job.etag,
            total_rows_count=row_iterator.total_rows,
            rows_count=len(rows),
            page_token=None,
            next_page_token=None,
            rows=rows
        )

def array_measure_generator_from_declarative(definition):
    generated_def = {}

    for key, value in definition.items():
        name_format = value['name_format']
        declarative_keys = value['keys']
        
        keys, values = zip(*declarative_keys.items())
        
        listed_keys = [
            dict(zip(keys, bundle)) for bundle in itertools.product(*values)
        ]

        generated_def[key] = {
            'name': name_format,
            'keys': listed_keys,
            'value_column': value.get('value_column', 'value_f')
        }

    return generated_def

def check_schema_has_field_name(schema, field_name, path=[]):
    field_name_parts = field_name.split('.')
    field_name = field_name_parts[0]
    other_parts = field_name_parts[1:]

    if field_name == '*':
        return True

    found_field = None
    for field in schema:
        if field.name == field_name:
            found_field = field
            break
    else:
        raise ValueError(f'Schema does not have a field named {".".join(path + [field_name])}')

    if found_field.field_type == 'RECORD':
        if other_parts:
            return check_schema_has_field_name(found_field.fields, '.'.join(other_parts), path=path+[field_name])
        else:
            return True # Asked for this record explicitely
    elif other_parts:
        raise ValueError(f'Sub field is not of type records but sub keys were specified.')
    else:
        return True


def spintop_filters_to_bq(filters, table_schema=None):
    OP_MAP = {
        'eq': _bq_simple_operator('='),
        'neq': _bq_simple_operator('!='),
        'lt': _bq_simple_operator('<'),
        'lte': _bq_simple_operator('<='),
        'gt': _bq_simple_operator('>'),
        'gte': _bq_simple_operator('>='),
        'match': None,
        'search': bq_regex_search
    }

    bq_filters = []
    params = {}

    for f in filters:
        if table_schema:
            check_schema_has_field_name(table_schema, f.name)
        
        # Creates an md5 hash of this string to use as an identifer for the @x notation of BigQuery parameters.
        internal_name = safe_identifier(f.name + '/' + f.op)

        field_type = f.field_type
        if field_type == 'datetime':
            field_type = 'timestamp'

        field_type = field_type.upper()
        params[internal_name] = (field_type, f.val)
        op = OP_MAP.get(f.op)

        bq_filters.append(op(f.name, field_type, internal_name))
    
    return bq_filters, params
            

def bq_regex_search(field_name, field_type, internal_name):
    return f'REGEXP_CONTAINS(CAST({field_name} AS {field_type}), CAST(@{internal_name} AS {field_type}))'

def _bq_simple_operator(op):
    return lambda field_name, field_type, internal_name: f'CAST({field_name} AS {field_type}) {op} CAST(@{internal_name} AS {field_type})'

def safe_identifier(field_name):
    result = hashlib.md5(field_name.encode())
    return '_' + result.hexdigest()

def safe_columns(columns, table_schema=None):

    for column in columns:
        if table_schema:
            check_schema_has_field_name(table_schema, column)

    return columns

def spintop_order_by_to_bq(order_bys, table_schema=None):

    bq_order_by = []

    for order_by in order_bys:
        if table_schema:
            check_schema_has_field_name(table_schema, order_by.name)
        
        asc_or_desc = 'ASC' if order_by.asc else 'DESC'
        bq_order_by.append(f'{order_by.name} {asc_or_desc}')
    
    return bq_order_by


import json

from flask.json import JSONEncoder as FlaskEncoder

from google.cloud import pubsub
from google.api_core.exceptions import NotFound, AlreadyExists

from spintop.models import get_json_serializer
from spintop.messages import SpintopMessagePublisher

from spintop.utils.encoding import JSONEncoder

def get_encoder():
    return JSONEncoder(encoder_cls=FlaskEncoder, sort_keys=True)

class GCPPubSub(SpintopMessagePublisher):
    def __init__(self, project_id):

        self.project_id = project_id
        self.publisher = pubsub.PublisherClient()
        self.encoder = get_encoder()

    def create_topic(self, topic_name):
        topic_path = self.topic_path(topic_name)
        self.publisher.create_topic(topic_path)

    def topic_path(self, topic_name):
        return self.publisher.topic_path(self.project_id, topic_name)

    def try_create_topic(self, topic_name):
        try:
            self.create_topic(topic_name)
        except AlreadyExists:
            pass

    def publish(self, topic_name, data, origin='spintop-api', auto_create=True):

        data_bytes = self.encoder.encode(data)

        topic_path = self.publisher.topic_path(self.project_id, topic_name)

        future = self.publisher.publish(
            topic_path, data_bytes, origin=origin
        )
        try:
            return future.result()
        except NotFound:
            if auto_create:
                self.create_topic(topic_name)
                return self.publish(topic_name, data, origin=origin, auto_create=False)
            else:
                raise
    
    def create_subscription(self, subscription_name, topic_name):
        subscriber = pubsub.SubscriberClient()
        subscription_path = subscriber.subscription_path(
            self.project_id, subscription_name
        )
        topic_path = self.publisher.topic_path(self.project_id, topic_name)
        subscriber.create_subscription(name=subscription_path, topic=topic_path)

    def try_create_subscription(self, subscription_name, topic_name):
        try:
            self.create_subscription(subscription_name, topic_name)
        except AlreadyExists:
            pass

    def pull_subscription(self, subscription_name, callback, max_messages_at_a_time=1):
        subscriber = pubsub.SubscriberClient()
        subscription_path = subscriber.subscription_path(
            self.project_id, subscription_name
        )
        
        def callback_wrapper(encoded_message):
            message = self.encoder.decode(encoded_message.data)
            callback(message)
            encoded_message.ack()

        flow_control = pubsub.types.FlowControl(max_messages=max_messages_at_a_time)
        streaming_pull_future = subscriber.subscribe(
            subscription_path, callback=callback_wrapper, flow_control=flow_control
        )
        # try:
        # When `timeout` is not set, result() will block indefinitely,
        # unless an exception is encountered first.
        streaming_pull_future.result(timeout=100)
        # except:  # noqa
        #     streaming_pull_future.cancel()
        #     raise



from marshmallow import Schema, fields

class ManyRecordsSchema(Schema):
    records = fields.List(fields.Dict())
    
class RecordsCountSchema(Schema):
    count = fields.Int()
    
class TokenSchema(Schema):
    token = fields.String()

class DashboardSchema(Schema):
    base_url = fields.String()
    url = fields.String()
    name = fields.String()
    sharing_token = fields.String()
    sharing_secret = fields.String()

class TopicSchema(Schema):
    key = fields.String()
    name = fields.String()
    path = fields.String()

records_schema = ManyRecordsSchema()
record_count_schema = RecordsCountSchema()
token_schema = TokenSchema()

dashboards_schema = DashboardSchema(many=True)
topics_schema = TopicSchema(many=True)




#!/usr/bin/env python3

import argparse
import io
import sys
import simplejson as json
import logging
import collections
import threading
import http.client
import urllib
import pkg_resources


from jsonschema import validate
import singer

from oauth2client import tools
from tempfile import TemporaryFile

import google.auth
from google.cloud import bigquery
from google.cloud.bigquery import Dataset
from google.api_core import exceptions

from .base import BigQueryDictLoadJobBuilder
from .schema import build_schema

from spintop.analytics.singer import parse_message_dict

logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
logger = singer.get_logger()

SCOPES = ['https://www.googleapis.com/auth/bigquery','https://www.googleapis.com/auth/bigquery.insertdata']
CLIENT_SECRET_FILE = 'client_secret.json'
APPLICATION_NAME = 'Singer BigQuery Target'

StreamMeta = collections.namedtuple('StreamMeta', ['schema', 'key_properties', 'bookmark_properties'])

def emit_state(state):
    if state is not None:
        line = json.dumps(state)
        logger.debug('Emitting state {}'.format(line))
        sys.stdout.write("{}\n".format(line))
        sys.stdout.flush()

def clear_dict_hook(items):
    return {k: v if v is not None else '' for k, v in items}

def merge_schemas(*schemas):
    all_fields = collections.OrderedDict()
    for schema in schemas:
        for field in schema:
            key = field.name
            if key in all_fields and field != all_fields[key]:
                raise ValueError(f'Schema type conflict for field {key}: {field} vs {all_fields[key]}')
            all_fields[key] = field
    
    return list(all_fields.values())

class TargetBigQuery(object):
    def __init__(self, bigquery_client, dataset):
        self.bigquery_client = bigquery_client
        self.dataset = dataset

    @classmethod
    def from_project_dataset_id(cls, project_id, dataset_id, job_project_id=None):
        """
        Params:
            project_id: The project where the dataset resiedes.
            dataset_id: The dataset id
            job_project_id (optional): The project where queries will be executed. Uses project_id if not specified.
        """

        if job_project_id is None:
            job_project_id = project_id

        credentials, _ = google.auth.default(
            scopes=[
                "https://www.googleapis.com/auth/drive",
                "https://www.googleapis.com/auth/bigquery",
            ]
        )

        bigquery_client = bigquery.Client(credentials=credentials, project=job_project_id)
        dataset_ref = bigquery_client.dataset(dataset_id, project=project_id)
        dataset = Dataset(dataset_ref)
        return cls(bigquery_client=bigquery_client, dataset=dataset)

    def try_create_dataset(self):
        try:
            dataset = self.bigquery_client.create_dataset(self.dataset)
        except (exceptions.Conflict, exceptions.Forbidden):
            pass

    def persist_lines_job(self, lines=None, truncate=False, validate_records=True):
        messages = []
        for line in lines:
            key = idx # Unique JSON key for the schema possibly
            try:
                msg = singer.parse_message(line)
            except json.decoder.JSONDecodeError:
                logger.error("Unable to parse:\n{}".format(line))
                raise
            
            messages.append(msg)
        return persist_messages_job(messages, truncate=truncate, validate_records=validate_records)

    def persist_messages_job(self, messages=None, truncate=False, validate_records=True):
        state = None
        schemas = {}
        key_properties = {}
        tables = {}
        rows = collections.defaultdict(dict)
        errors = {}
        current_schema_key = None

        bigquery_client = self.bigquery_client
        dataset = self.dataset

        for idx, msg in enumerate(messages):
            key = idx # Unique JSON key for the schema possibly
            if not isinstance(msg, singer.Message):
                msg = parse_message_dict(msg)
            # try:
            #     msg = singer.parse_message(line)
            # except json.decoder.JSONDecodeError:
            #     logger.error("Unable to parse:\n{}".format(line))
            #     raise
            
            if not rows[msg.stream].get(current_schema_key):
                rows[msg.stream][current_schema_key] = BigQueryDictLoadJobBuilder(dataset.dataset_id, msg.stream, truncate=truncate, client=bigquery_client, dataset=dataset)

            job_builder = rows[msg.stream][current_schema_key]

            if isinstance(msg, singer.RecordMessage):
                # if not current_schema_key:
                #     raise Exception("A record for stream {} was encountered before a corresponding schema".format(msg.stream))


                if validate_records and current_schema_key:
                    schema = schemas[current_schema_key]
                    validate(msg.record, schema)

                job_builder.add_row(msg.record)

                state = None

            elif isinstance(msg, singer.StateMessage):
                logger.debug('Setting state to {}'.format(msg.value))
                state = msg.value

            elif isinstance(msg, singer.SchemaMessage):
                table = msg.stream 
                key_properties[table] = msg.key_properties
                
                table_ref = dataset.table(table)
                
                job_builder = rows[msg.stream][current_schema_key]
                current_schema = schemas.get(current_schema_key, None)
                
                if current_schema is None and not truncate:
                    # Get or create from big query
                    try:
                        remote_table = bigquery_client.get_table(table_ref)
                    except exceptions.NotFound:
                        remote_table = bigquery_client.create_table(table_ref)

                    current_schema = remote_table.schema
                    

                if key not in schemas:
                    new_schema = build_schema(msg.schema)
                    if not truncate:
                        merged_schema = merge_schemas(current_schema, new_schema)
                    else:
                        merged_schema = new_schema
                    schemas[key] = merged_schema
                    tables[table] = bigquery.Table(dataset.table(table), schema=merged_schema)

                current_schema_key = key

            elif isinstance(msg, singer.ActivateVersionMessage):
                # This is experimental and won't be used yet
                pass

            else:
                raise Exception("Unrecognized message {}".format(msg))
        
        jobs = []
        for table in rows.keys():
            tables_schemas = rows[table]
            for schema_key in tables_schemas.keys():
                job_builder = tables_schemas[schema_key]
                if not job_builder.is_empty:
                    job_builder.schema = schemas.get(schema_key, None)

                    load_job = job_builder.build_job()
                    jobs.append(load_job)

        for job in jobs:
            result = job.result()
            # logger.info(result)

        # for table in errors.keys():
        #     if not errors[table]:
        #         print('Loaded {} row(s) into {}:{}'.format(rows[table], dataset_id, table), tables[table].path)
        #     else:
        #         print('Errors:', errors[table], sep=" ")

        return state

    def persist_lines_stream(self, project_id, dataset_id, lines=None, validate_records=True):
        raise NotImplementedError()
        state = None
        schemas = {}
        key_properties = {}
        tables = {}
        rows = {}
        errors = {}

        bigquery_client = bigquery.Client(project=project_id)

        dataset_ref = bigquery_client.dataset(dataset_id)
        dataset = Dataset(dataset_ref)
        try:
            dataset = bigquery_client.create_dataset(Dataset(dataset_ref)) or Dataset(dataset_ref)
        except exceptions.Conflict:
            pass

        for line in lines:
            try:
                msg = singer.parse_message(line)
            except json.decoder.JSONDecodeError:
                logger.error("Unable to parse:\n{}".format(line))
                raise

            if isinstance(msg, singer.RecordMessage):
                if msg.stream not in schemas:
                    raise Exception("A record for stream {} was encountered before a corresponding schema".format(msg.stream))

                schema = schemas[msg.stream]

                if validate_records:
                    validate(msg.record, schema)

                errors[msg.stream] = bigquery_client.insert_rows_json(tables[msg.stream], [msg.record])
                rows[msg.stream] += 1

                state = None

            elif isinstance(msg, singer.StateMessage):
                logger.debug('Setting state to {}'.format(msg.value))
                state = msg.value

            elif isinstance(msg, singer.SchemaMessage):
                table = msg.stream 
                schemas[table] = msg.schema
                key_properties[table] = msg.key_properties
                tables[table] = bigquery.Table(dataset.table(table), schema=build_schema(schemas[table]))
                rows[table] = 0
                errors[table] = None
                try:
                    tables[table] = bigquery_client.create_table(tables[table])
                except exceptions.Conflict:
                    pass

            elif isinstance(msg, singer.ActivateVersionMessage):
                # This is experimental and won't be used yet
                pass

            else:
                raise Exception("Unrecognized message {}".format(msg))

        for table in errors.keys():
            if not errors[table]:
                logging.info('Loaded {} row(s) into {}:{}'.format(rows[table], dataset_id, table, tables[table].path))
                emit_state(state)
            else:
                logging.error('Errors:', errors[table])

        return state

def collect():
    try:
        version = pkg_resources.get_distribution('target-bigquery').version
        conn = http.client.HTTPConnection('collector.singer.io', timeout=10)
        conn.connect()
        params = {
            'e': 'se',
            'aid': 'singer',
            'se_ca': 'target-bigquery',
            'se_ac': 'open',
            'se_la': version,
        }
        conn.request('GET', '/i?' + urllib.parse.urlencode(params))
        conn.getresponse()
        conn.close()
    except:
        logger.debug('Collection request failed')

def main():
    try:
        parser = argparse.ArgumentParser(parents=[tools.argparser])
        parser.add_argument('-c', '--config', help='Config file', required=True)
        flags = parser.parse_args()

    except ImportError:
        flags = None

    with open(flags.config) as input:
        config = json.load(input)

    if not config.get('disable_collection', False):
        logger.info('Sending version information to stitchdata.com. ' +
                    'To disable sending anonymous usage data, set ' +
                    'the config parameter "disable_collection" to true')
        threading.Thread(target=collect).start()

    if config.get('replication_method') == 'FULL_TABLE':
        truncate = True
    else:
        truncate = False

    validate_records = config.get('validate_records', True)

    input = io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8')

    if config.get('stream_data', True):
        state = persist_lines_stream(config['project_id'], config['dataset_id'], input, validate_records=validate_records)
    else:
        state = persist_lines_job(config['project_id'], config['dataset_id'], input, truncate=truncate, validate_records=validate_records)

    emit_state(state)
    logger.debug("Exiting normally")


if __name__ == '__main__':
    main()


from datetime import datetime

from spintop.persistence.base import PersistenceFacade

from ..models import PersistenceRecordCollection, Query, get_json_serializer, SerializedPersistenceRecord

from .schemas import records_schema, record_count_schema

from ..logs import _logger

logger = _logger('spintop-api.records')

class SpintopAPIPersistenceFacade(PersistenceFacade):
    def __init__(self, spintop_api=None, uri=None, database_name=None, message_publisher=None, env=None, **unused):
        
        if spintop_api is None:
            spintop_api = default_spintop_api(env, uri, database_name)

        self.spintop_api = spintop_api
        self.serializer = get_json_serializer()
        spintop_api.register_persistence_facade(self)
        super().__init__(self.serializer, messages=message_publisher)

    @property
    def session(self):
        return self.spintop_api.session

    def _create(self, records):
        logger.debug(f'Creating records at {self.spintop_api.api_url}')
        serialized = self._records_in_schema(records)
        return self.session.post(self.spintop_api.get_link('records.create'), json=serialized)
        
    def _records_in_schema(self, records):
        return records_schema.dump({'records': [record.as_dict() for record in records]})

    def _retrieve(self, query, limit_range=None, with_data=True):
        logger.debug(f'Retrieving records from {self.spintop_api.api_url}')
        if limit_range is None:
            limit_range = (0,500)
        query_dict = query.as_dict()
        if limit_range:
            query_dict['limit_range_inf'] = limit_range[0]
            query_dict['limit_range_sup'] = limit_range[1]
            
        query_dict['with_data'] = with_data

        resp = self.session.post(self.spintop_api.get_link('records.search'), json=query_dict)
        records = records_schema.load(resp.json())['records']
        return (SerializedPersistenceRecord(record) for record in records)
        
    def _count(self, query):
        query_dict = query.as_dict()
        resp = self.session.post(self.spintop_api.get_link('records.count'), json=query_dict)
        return record_count_schema.load(resp.json())['count']

    def _update(self, records):
        logger.debug(f'Updating records at {self.spintop_api.api_url}')
        serialized = self._records_in_schema(records)
        return self.session.put(self.spintop_api.get_link('records.update'), json=serialized)
    
    def _delete(self, query):
        logger.debug(f'Deleting records from {self.spintop_api.api_url}')
        query_dict = query.as_dict()
        return self.session.delete(self.spintop_api.get_link('records.delete'), json=query_dict)
        
    def __str__(self):
        return f'{self.__class__.__name__}(url={self.spintop_api.api_url})'

def default_spintop_api(env, uri=None, database_name=None):
    if uri is not None and database_name is not None:
        return env.spintop_factory(org_id=database_name)
    else:
        raise ValueError('uri + database_name parameters are required.')


from typing import List, Union, Any, Dict

from spintop.models import BaseDataClass, get_serializer, FilterSpec

class DimensionSpec(BaseDataClass):
    name: str
    field_type: str
    label: str = None
    description: str = None

class TableSpec(BaseDataClass):
    filters: List[FilterSpec] = list
    custom_dimensions: List[DimensionSpec] = list

class EnvDataspecs(BaseDataClass):
    steps: TableSpec = None
    production_nodes: TableSpec = None
    project_id: str = None
    dataset_id: str = None

class OrgDataEnv(BaseDataClass):
    name: str
    vars: dict = dict
    dataspecs: EnvDataspecs = None

class Dashboard(BaseDataClass):
    name: str
    base_url: str = None
    sharing_secret: str = None
    sharing_token: str = None
    url: str = None

class Organization(BaseDataClass):
    key: str
    project_id: str
    default_env: str = None
    dashboards: List[Dashboard] = list
    envs: List[OrgDataEnv] = list

    def find_env(self, env_name):
        for env in self.envs:
            if env.name == env_name:
                return env
        else:
            raise ValueError(f'Org {self.key} has no env named {env_name}')

    def env_dataset_name(self, env_name):
        return f'data__{self.key}__{env_name}'

class ManyOrganizations(BaseDataClass):
    orgs: List[Organization]

    def __iter__(self):
        return iter(self.orgs)
    
    def find_org(self, org_key):
        for org in self.orgs:
            if org.key == org_key:
                return org
        else:
            raise ValueError(f'No org named {org_key}')
    

class User(BaseDataClass):
    username: str = None
    user_id: str
    scope: List[str]
    permissions: List[str]
    user_type: str
    organizations: List[Organization]

class UnitMatch(BaseDataClass):
    lc_type: str
    match: str



import yaml


class IntegrationConfig(object):
    def __init__(self, username, password, expected_org):
        self.username = username
        self.password = password
        self.expected_org = expected_org
        self.spintop_api_url = None
        
    def login_and_assert(self, auth_module):
        self.login(auth_module)
        self.assert_credentials_good(auth_module.credentials)
        self.assert_org_in_users(auth_module)

    def assert_credentials_good(self, credentials):
        assert credentials['username'] == self.username

    def assert_credentials_empty(self, credentials):
        assert not credentials

    def login(self, auth_module):
        auth_module.login_user_pass(self.username, self.password)

    def assert_org_in_users(self, auth_module):
        assert self.expected_org in auth_module.user_orgs

    def assert_org_match(self, org):
        assert org['key'] == self.expected_org

    @classmethod
    def load_file(cls, filepath):
        with open(filepath) as conffile:
            return cls(**yaml.load(conffile, Loader=yaml.FullLoader))
        


from spintop.models import VersionIDRecord, FeatureRecord
from spintop.models.persistence import ProcessStepIDRecord

from spintop.models import get_json_serializer

from .features import MeasureFeatureRecord, PhaseFeatureRecord, LCFeature
from .files import SpintopFileReference
from .steps import LCStep, LCUnit, LCTypes 

feature_fields = FeatureRecord.fields()
index_field = feature_fields.index

def record_to_lifecycle_step(record, pipeline_uuid=None, filename=None):

    step = cast_to(record, LCStep, units=[])

    add_unit_if_exists(step, record.dut, LCTypes.DUT)
    add_unit_if_exists(step, record.step, LCTypes.STEP)
    add_unit_if_exists(step, record.station, LCTypes.STATION)
    add_unit_if_exists(step, VersionIDRecord.create(record.operator), LCTypes.OPERATOR)
    step.pipeline_uuid = pipeline_uuid
    step.filename = filename
    return step

def add_unit_if_exists(step, version_id, lc_type):
    if version_id:
        step.units.append(LCUnit.from_version_id(version_id, lc_type))

def cast_to(record, cls, **others):
    cls_fields = set(field.field_name_ for field in cls.fields())

    data = {
        field.field_name_: field(record) for field in record.fields() if field.field_name_ in cls_fields
    }
    data.update(others)

    return cls(
        **data
    )

def feature_to_lifecycle_feature(record_uuid, feature, pipeline_uuid=None):
    feature_uuid = f'{record_uuid}-feat{index_field(feature)}'

    return cast_to(feature, LCFeature,  uuid=feature_uuid, record_uuid=record_uuid, pipeline_uuid=pipeline_uuid)


STREAM_SUFFIX = '_stream'

class LifecycleStream(object):
    def __init__(self, table_name, model_type, analytics, org_data):
        self.name = table_name
        self.stream_name = table_name + STREAM_SUFFIX
        self.model_type = model_type

        self.analytics = analytics
        self.org_data = org_data

        self.reserve()

    def update(self, raw_data):
        return self.analytics.update_named_stream(self.name, raw_data, stream_type=self.model_type)

    def singer_factory(self):
        return self.analytics.singer_factory(self.stream_name, record_cls=self.model_type)

    def on_stream_update(self, dataset_fns):
        pass
        # query = dataset_fns.map_stream_to_latest_views(self.stream_name)
        # # Virtual view
        # query.update_view(self.name)

    def reserve(self):
        self.analytics.reserve_stream_name(self.name, self.on_stream_update, stream_type=self.model_type, stream_name=self.stream_name)

class StepsLifecycle(LifecycleStream):

    def on_stream_update(self, dataset_fns):
        super().on_stream_update(dataset_fns)
        # Create dut states view table
        # org_config = self.org_data

        # table_name = self.name + '_dut'
        # config = org_config.get(table_name, {})

        # query = dataset_fns.lc_unit_yields(self.name, LCTypes.DUT, **config)
        # # Virtual view
        # query.update_view(table_name)

class FeaturesLifecycle(LifecycleStream):
    
    def on_stream_update(self, dataset_fns):
        super().on_stream_update(dataset_fns)
        # Create dut states view table
        org_config = self.org_data

        config = org_config.get('measures_marts', {})

        table_name = self.name + '_mart'

        for table_suffix, measures in config.items():
            mart_table_name = table_name + table_suffix
            query = dataset_fns.measure_values_mart(self.name, measure_names=measures)
            # Virtual view
            query.update_view(mart_table_name)

def reserve_lifecycle_streams(analytics, env):
    analytics.steps = StepsLifecycle('steps', LCStep, analytics, env['SPINTOP_DA_LC_STEPS'])
    analytics.features = FeaturesLifecycle('features', LCFeature, analytics, env['SPINTOP_DA_LC_FEATURES'])
    analytics.processed_files = LifecycleStream('files', SpintopFileReference, analytics, None)




from .operations import MongoOperations, ASCENDING
from .serialization import DataClassSerializer

from spintop.models import (
    TestIDRecord,
    SpintopTestRecord,
    SpintopTestRecordBuilder,
    Query,
    FieldsOf
)


from ...logs import _logger


SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES = {}

SIMPLE_MAPPER_DATACLASSES_INIT_FNS = {}

logger = _logger('mongo')

def register_simple_mapper(cls, collection_name):
    SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES[cls] = collection_name
    SIMPLE_MAPPER_DATACLASSES_INIT_FNS[cls] = None
    def _register_init(fn):
        SIMPLE_MAPPER_DATACLASSES_INIT_FNS[cls] = fn
        return fn
    return _register_init

TestUUID = FieldsOf(TestIDRecord).uuid
DateTime = FieldsOf(TestIDRecord).start_datetime
Index = FieldsOf(TestIDRecord).index

# Mappers
@register_simple_mapper(TestIDRecord, 'test-features')
def init_test_record_mapper(mapper):
    test_uuid_field = TestUUID.name_
    datetime_field = DateTime.name_
    index_field = Index.name_

    mapper.ops.create_index([(test_uuid_field, ASCENDING), (index_field, ASCENDING)], unique=True)
    mapper.ops.grouped_by = test_uuid_field

def create_mappers(mongo_db, serializer=None, operations_factory=None):
    
    # Simple mappers are 1:1 with a document in a mongo collection.
    mappers = {}
    for dataclass, collection_name in SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES.items():
        init_fn = SIMPLE_MAPPER_DATACLASSES_INIT_FNS[dataclass]
        mappers[dataclass] = simple_mongo_mapper_factory(
            mongo_db, 
            collection_name, 
            serializer=serializer,
            on_init=init_fn,
            operations_factory=operations_factory
        )
    
    return mappers

def simple_mongo_mapper_factory(mongo_db, collection_name, serializer=None, on_init=None, operations_factory=None):
    if operations_factory is None: operations_factory = MongoOperations

    ops = operations_factory(mongo_db[collection_name])
    return SimpleMongoMapper(ops, serializer=serializer, on_init=on_init, name=collection_name)


class SimpleMongoMapper(object):
    def __init__(self, ops, serializer=None, on_init=None, name="simple"):
        if serializer is None: serializer = DataClassSerializer()
        self.ops = ops
        self.dc_serializer = serializer
        self.on_init = on_init
        self.logger = logger.getChild(name)
    
    def init(self):
        if self.on_init: self.on_init(self)
    
    def create(self, serialized_dcs):
        serialized_dcs = [self.dc_serializer.serialized_to_mongo_id(dc, keep_id_if_none=False) for dc in serialized_dcs]
        
        resp = self.ops.insert_many(serialized_dcs)
        self.logger.info('Inserted %d new documents' % len(resp.inserted_ids))
        
        # ids are inserted client side into the 'serialized' dictionnary.
        # We can re-insert _ids into dataclasses.
        for dc, serialized in zip(serialized_dcs, serialized_dcs):
            dc['oid'] = serialized['_id']
            
    def count(self, query=Query()):
        return self.ops.count(query)

    def count_total(self, query=Query()):
        return self.ops.count_total(query)

    def collection_stats(self):
        return self.ops.collection_stats()

    def retrieve(self, query=Query(), limit_range=None):
        responses = self.ops.find(query, limit_range=limit_range)
        return [self.dc_serializer.serialized_from_mongo_id(dc) for dc in responses]
        
    def update(self, serialized_dcs, query_fields):
        """Non-working. Facade uses a combination of delete + create to mimick update behavior."""
        # serialized_dcs = self.dc_serializer.serialize_many(dataclasses)
        if not query_fields:
            raise ValueError('query_fields must be a non-empty list of MetaFields.')

        def obj_query(obj):
            q = {field.name_: field.value_of_serialized(obj) for field in query_fields}
            return q

        self.ops.update_many(serialized_dcs, obj_query)
        
    def delete(self, query=Query()):
        resp = self.ops.delete_many(query)
        self.logger.info('Deleted %d documents' % resp.deleted_count)


from typing import Optional
from collections.abc import Mapping

from .base import BaseDataClass, model_property

NO_VERSION = 'none'

class VersionIDRecord(BaseDataClass):
    id: Optional[str]
    version: Optional[str]
    
    @classmethod
    def create(cls, id_or_dict, version=None):
        if isinstance(id_or_dict, cls):
            return id_or_dict
        if isinstance(id_or_dict, Mapping):
            id = id_or_dict.get('id', None)
            version = id_or_dict.get('version', None)
        else:
            id = id_or_dict
            
        if version is None:
            version = NO_VERSION
            
        return cls(
            id=id,
            version=version
        )
        
    def match(self, other):
        id_match = self.id == other.id if self.id is not None else True
        version_match = self.version == other.version if self.version is not None else True
        return id_match and version_match

    def __str__(self):
        if self.version and self.version != NO_VERSION:
            return f'{self.id}-{self.version}'
        elif self.id:
            return str(self.id)
        else:
            return 'default'

TestbenchIDRecord = VersionIDRecord

DutIDRecord = VersionIDRecord

class OutcomeData(BaseDataClass):
    message: str = ''
    is_pass: bool = True
    state: str = ''

    @property
    def is_skip(self):
        return self._get_bool_state_prop('skip')

    @is_skip.setter
    def is_skip(self, value):
        return self._set_bool_state_prop('skip', value)

    @property
    def is_abort(self):
        return self._get_bool_state_prop('abort')

    @is_abort.setter
    def is_abort(self, value):
        return self._set_bool_state_prop('abort', value)
        
    @property
    def is_ignore(self):
        return self._get_bool_state_prop('ignore')

    @is_ignore.setter
    def is_ignore(self, value):
        return self._set_bool_state_prop('ignore', value)

    def _get_bool_state_prop(self, state_str):
        return self.state == state_str
    
    def _set_bool_state_prop(self, state_str, value):
        if value:
            self.state = state_str
        else:
            self.state = ''

    def __bool__(self):
        return self.is_pass and not self.is_abort

    @classmethod
    def create(cls, outcome):
        if outcome is None:
            outcome = True
            
        if isinstance(outcome, cls):
            pass # already correct type
        elif isinstance(outcome, bool):
            outcome = cls(is_pass=outcome)
        elif isinstance(outcome, Mapping):
            outcome = cls(**outcome)
        else:
            raise TypeError((
    'Outcome (%s) should either be a boolean that indicates if this test passed or failed, ' 
    'or a dictionnary of one or more of the following attributes: %s'
            ) % (outcome, [f.name for f in cls.dataclass_fields()]))
        return outcome

    def impose_upon(self, other_outcome):
        if not self.is_ignore:
            # Pass propagates by falseness. False will propagate up, not True
            other_outcome.is_pass = other_outcome.is_pass and self.is_pass

            # Abort propagates by trueness. True will propagate up, not False
            other_outcome.is_abort = other_outcome.is_abort or self.is_abort

            # Abort propagates by trueness. True will propagate up, not False
            other_outcome.is_skip = other_outcome.is_skip or self.is_skip



from .get_recursive import (
    GetRecursiveMixin,
    GetRecursiveKeyMixin,
    AnonymousGetRecursive,
    GetRecursive,
    GetRecursiveAccessor
)

from .misc import *

from .requests_utils import spintop_session


from .meta import FieldsOf
from .test_records import MeasureFeatureRecord

def value_of_feature_named(name):
    measure_fields = FieldsOf(MeasureFeatureRecord)
    name_field = measure_fields.name.name_
    value_fields = [
        measure_fields.value_f.name_,
        measure_fields.value_s.name_,
    ]
    def _accessor(feat):
        if feat.get_recursive(name_field) != name:
            raise AttributeError('Wrong feature')
        else:
            for value_field in value_fields:
                value = feat.get_recursive(value_field)
                if value is not None:
                    break
            return value
    return _accessor


""" Storage-friendly internal data classes. """
import numbers

from collections import defaultdict
from collections.abc import Mapping
from base64 import urlsafe_b64decode, urlsafe_b64encode
from uuid import UUID, uuid4

from datetime import datetime
from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field, asdict
from typing import Union, List, Optional, Dict

from ..serialization import cls_of_obj

from ..base import (
    BaseDataClass, 
    _fields_cache,
    model_property
)

from ..common import DutIDRecord, TestbenchIDRecord, OutcomeData

from ..persistence import QualityStepIDRecord, uuid_generator

from ..view import ComplexPrimitiveView

from spintop.utils import isnan, utcnow_aware

NAME_SEPARATOR = ':'

def compute_stats(features):
    name_tuple_to_count_map = defaultdict(int)

    for feature in features:
        # For feature ('x', 'y', 'z'), increment ('x',), ('x', 'y') and ('x', 'y', 'z')
        name_tuple = feature.name_tuple

        for i in range(1, len(name_tuple)):
            sub_name = name_tuple[:i]
            name_tuple_to_count_map[sub_name] += 1

    for feature in features:
        feature.feature_count = name_tuple_to_count_map[feature.name_tuple]


class FeatureRecordNoData(BaseDataClass):
    """ Feature ID is determined by its name.
    """
    oid: Optional[str]
    name: Optional[str] = ''
    description: str = ''
    version: int = 0
    depth: int = 0
    index: int = 0
    ancestors: List[str] = list
    original: bool = True
    
    @classmethod
    def defaults(cls, **others):
        others['name'] = others.get('name', '')
        others['ancestors'] = others.get('ancestors', [])
        return super().defaults(**others)

    @classmethod
    def data_dataclass_fields(cls):
        """ The data fields are all fields except the ones declare up to this cls
        in the cls MRO.
        
        If a subclass defines only a dataclass field 'foo', this method will return
        foo only as a data field.
        """
        return set(_fields_cache.retrieve(cls)) - set(_fields_cache.retrieve(FeatureRecordNoData))
    
    @property
    def name_tuple(self):
        return tuple(self.ancestors) + (self.name,)

    @property
    def complete_name(self) -> str:
        return NAME_SEPARATOR.join(self.name_tuple)
    
class FeatureRecord(FeatureRecordNoData):
    # user_data: Dict[str, Optional[Any]] = dict 
    user_data: dict = dict
    outcome: OutcomeData = OutcomeData
    feature_count: int = 0 # Number of children
    
    def __post_init__(self):
        super().__post_init__()
        self.outcome = OutcomeData.create(self.outcome)

    @property
    def total_feature_count(self):
        return self.feature_count

    @classmethod
    def defaults(cls, **others):
        others['user_data'] = others.get('user_data', {})
        others['feature_count'] = others.get('feature_count', 0)
        return super(FeatureRecord, cls).defaults(**others)
    
class PhaseFeatureRecord(FeatureRecord):
    duration: Optional[float] = None
    
class TestIDRecord(QualityStepIDRecord, PhaseFeatureRecord):
    testbench: TestbenchIDRecord = model_property()

    @testbench.getter
    def testbench(self):
        return self.step

    @testbench.setter
    def testbench(self, value):
        self.step = value

    @property
    def testbench_name(self) -> Optional[str]:
        return self.testbench.id

    @testbench_name.setter
    def testbench_name(self, value):
        self.testbench.id = value


TestRecordSummary = TestIDRecord # backward compat

class MeasureFeatureRecord(FeatureRecord):
    value_f: Optional[float]
    value_s: Optional[str]

    @property
    def value(self) -> Union[float, str, None]:
        if self.value_f is not None:
            return self.value_f
        else:
            return self.value_s

    @value.setter
    def value(self, value):
        if value is None or isinstance(value, str):
            self.value_f = None
            self.value_s = value
        elif isinstance(value, numbers.Number):
            self.value_f = value
            self.value_s = None
        else:
            raise ValueError('Only string or numeric types are supported. Received {!r} of type {}'.format(value, type(value)))
    
class DutOp(BaseDataClass):
    dut_match: DutIDRecord
    dut_after: DutIDRecord
    op_datetime: datetime = utcnow_aware
    
    @classmethod
    def create(cls, dut_match, dut_after, op_datetime):
        return cls(
            dut_match=DutIDRecord.create(dut_match),
            dut_after=DutIDRecord.create(dut_after),
            op_datetime=op_datetime
        )
    
    def does_applies_to(self, dut, on_datetime):
        return self.dut_match.match(dut) and self.op_datetime < on_datetime
    
    def apply_or_return(self, dut, on_datetime):
        if self.does_applies_to(dut, on_datetime):
            return self.apply(dut)
        else:
            return dut
        
    def apply(self, dut):
        """Apply op to dut. Does not check for datetime. """
        new_dut = self.dut_after.copy()
        if new_dut.id is None:
            new_dut.id = dut.id
        return new_dut
        


import re
import datetime as dt
from time import time
from functools import lru_cache
from collections.abc import Mapping
import json

from copy import deepcopy
from contextlib import contextmanager

from spintop.models.serialization import get_serializer, get_json_serializer
from spintop.utils import utcnow_aware, dict_ops

from .models import AnalyticsResponse
from . import singer
from ..logs import _logger

logger = _logger('singer')

DATETIME_TYPE = {'type': 'string', 'format': 'date-time'}

REQUEST_ID = '_spintop_request_id'

SCHEMA_ADDITIONS = {
    singer.SEQUENCE : {"type": "integer"},
    singer.BATCHED_AT: DATETIME_TYPE,
    REQUEST_ID: {"type": "string"}
}

@lru_cache()
def utcdatetime(_currenttime):
    """Computed only if current time changes."""
    return utcnow_aware()

class SingerMessagesFactory(object):
    def __init__(self, stream_name, add_null_to_fields=True):
        self.stream_name = stream_name
        self.transform_schema_kwargs = dict(
            add_null_to_fields=add_null_to_fields
        )
        self.logger = logger.getChild(stream_name)

    def get_default_schema(self):
        return None

    def schema(self, schema , key_properties):
        schema = deepcopy(schema)
        schema['properties'].update(SCHEMA_ADDITIONS)
        _schema_transform(schema, **self.transform_schema_kwargs)
        schema = change_keys_deep(schema, sanitize_key)
        return {
            'type': 'SCHEMA',
            'stream': self.stream_name,
            'key_properties': key_properties,
            'schema': schema
        }

    def record(self, data):
        current_time = sdc_sequence_now()
        data.update({
            singer.SEQUENCE: current_time,
            singer.BATCHED_AT: utcdatetime(current_time).isoformat() # serialize right away
        })

        data = change_keys_deep(data, sanitize_key, remove_none)
        return {
            'type': 'RECORD',
            'stream': self.stream_name,
            'record': data
        }

def sdc_sequence_now():
    return int(time()*1000*1000)

class ModelSingerMessagesFactory(SingerMessagesFactory):
    def __init__(self, stream_name, record_cls, serializer, **kwargs):
        self.record_cls = record_cls
        self.serializer = serializer
        super().__init__(stream_name, **kwargs)

    def get_default_schema(self):
        jsonschema = self.record_cls.dump_json_schema()
        return super().schema(jsonschema, key_properties=[])

    def schema(self, *args, **kwargs):
        raise TypeError('Model based message factory cannot receive a schema.')

    def record(self, data):
        """Data should be of type self.record_cls"""
        serialized = self.serializer.serialize(data, self.record_cls)
        return super().record(serialized)

class _NamedEndpointAccessor(object):
    def __init__(self, name):
        self.name = name
    
    def __get__(self, obj, cls):
        return NamedEndpoint(
            name=self.name,
            analytics=obj
        )

class NamedEndpoint(object):
    def __init__(self, name, analytics):
        self.name = name
        self.analytics = analytics
    
    def singer_factory(self):
        return self.analytics.singer_factory(self.name)

    def update(self, data):
        return self.analytics.update_named_stream(self.name, data)

class AbstractAnalyticsTarget(object):
    add_null_to_fields = False

    steps = _NamedEndpointAccessor('steps')
    features = _NamedEndpointAccessor('features')
    
    def __init__(self):
        self.serializer = get_serializer('tabular')
    
    def get_endpoint(self, name):
        return getattr(self, name)

    def update_named_stream(self, name, records, stream_type=None):
        with self.stream(name, stream_type) as batch:
            for record in records:
                batch.record(record)

    @property
    def functions(self):
        return None # To be overwritten

    @contextmanager
    def stream(self, stream_name, record_cls=None):
        batch = self.begin_batch(stream_name, record_cls=record_cls)
        yield batch
        batch.result = self.flush_batch(batch)

    def begin_batch(self, stream_name, record_cls=None):
        factory = self.singer_factory(stream_name, record_cls)
        return self.collect_factory(factory)

    def collect_factory(self, factory):
        return CollectMessagesFromFactory(factory)

    def singer_factory(self, stream_name, record_cls=None):
        if record_cls:
            return ModelSingerMessagesFactory(
                stream_name,
                record_cls,
                serializer=self.serializer,
                add_null_to_fields=self.add_null_to_fields
            )
        else:
            return SingerMessagesFactory(
                stream_name,
                add_null_to_fields=self.add_null_to_fields
            )

    def flush_batch(self, batch):
        raise NotImplementedError()
    
    def complete_response(self, response, request_id, stream_name):
        if request_id and not response.request_id:
            response.request_id = request_id

        if stream_name and not response.stream_name:
            response.stream_name = stream_name
        
        return response

class AbstractSingerTarget(AbstractAnalyticsTarget):
    
    def flush_batch(self, batch):
        singer_messages = batch.records
        if batch.schema:
            singer_messages.insert(0, batch.schema)

        batch.result = self.send_messages_dict(singer_messages, stream_name=batch.stream_name)

    def send_messages_dict(self, messages_dict, request_id=None, stream_name=None):
        return self.send_messages(self.json_dumps_messages(messages_dict), stream_name=stream_name)

    def send_messages(self, messages_str, request_id=None, stream_name=None):
        raise NotImplementedError()

    def json_dumps_messages(self, messages):
        serialized_messages = [self.serializer.serialize(msg) for msg in messages]
        return [json.dumps(ser_msg) for ser_msg in serialized_messages]

class CollectMessagesFromFactory(object):
    def __init__(self, factory):
        self.schema = factory.get_default_schema()
        self.records = []
        self.result = None
        self.factory = factory

    @property
    def stream_name(self):
        return self.factory.stream_name

    def schema(self, *args, **kwargs):
        if self.schema:
            raise ValueError('Message collector only supports one schema per batch.')
        
        self.schema = self.factory.schema(*args, **kwargs)
        return self.schema

    def record(self, *args, **kwargs):
        record = self.factory.record(*args, **kwargs)
        self.records.append(record)
        return record

### Replace datetime by string type.
_FIELD_TRANSFORM = {
    'datetime': lambda field: DATETIME_TYPE
}

def _schema_transform(schema, add_null_to_fields=True):
    try:
        fields = schema.get('properties', {})
    except AttributeError:
        return 
    
    for key, field in fields.items():
        # Try to get type in map, else keep same.
        # Also add null allowed everywhere.
        field.pop('default', None)
        most_important_field_type = get_field_type(field['type'])

        transformer = _FIELD_TRANSFORM.get(most_important_field_type, None)
        if transformer:
            update = transformer(field)
            field.update(update)

        if add_null_to_fields:
            field['type'] = [most_important_field_type, 'null']

        # in case it contains other properties (object)
        _schema_transform(field, add_null_to_fields=add_null_to_fields)

def get_field_type(field_type):
    if isinstance(field_type, str):
        return field_type
    elif field_type:
        for possible_type in field_type:
            if possible_type.lower() != 'null':
                return possible_type
    
    # Else
    return None


class RemoveValue(Exception):
    pass

def sanitize_key(key):
    return key.replace('.', '_')

def remove_none(value):
    if value is None:
        raise RemoveValue()
    else:
        return value

def noop(value):
    return value

def change_keys_deep(dict_obj, key_op, value_op=noop):
    new_flat = {}
    flat = dict_ops.flatten_dict(dict_obj)
    for key_tuple, value in flat.items():
        new_key_tuple = tuple(key_op(key) for key in key_tuple)
        try:
            new_flat[new_key_tuple] = value_op(value)
        except RemoveValue:
            pass

    return dict_ops.deepen_dict(new_flat)

DATETIME_REGEX = re.compile(r'(?:\D|^)(\d{4})-?([0-1]\d)-?([0-3]\d)(?:\D|$)')

def get_uuid_datetime(uuid):
    match = DATETIME_REGEX.search(uuid)
    if not match:
        raise ValueError('uuid does not contain a date in ISO format yyyymmdd or yyyy-mm-dd')
    
    try:
        year = int(match.group(1))
        month = int(match.group(2))
        day = int(match.group(3))
    except ValueError as e:
        raise ValueError(f'Unable to cast string to int: {str(e)}')

    try:
        return dt.datetime(year=year, month=month, day=day)
    except ValueError as e:
        raise ValueError(f'{year}-{month}-{day} is an invalid date: {str(e)}')


""" This is the 'meta' config module. Based on environ, decide on the best config object, param provider, and deployment type."""
import os

from incremental_module_loader import IncrementalModuleLoader

from spintop import __version__ as spintop_version

from .logs import _logger

try:
    from opencensus.ext.stackdriver import trace_exporter as stackdriver_exporter
    def initialize_gcp_tracer_export(project_id):
        exporter = stackdriver_exporter.StackdriverExporter(
            project_id=project_id
        )
        return exporter
except ImportError:
    initialize_gcp_tracer_export = lambda: None

logger = _logger('config-factory')

def no_provider_bootstrap(env, _modules=None):
    from ..services.param_provider import EnvOnlyParamProvider
    _modules = _common_bootstrap(env, _modules)
    _modules.load(param_provider=EnvOnlyParamProvider)
    return _modules

def gcp_bootstrap(env, _modules=None):
    _modules = _common_bootstrap(env, _modules)
    _modules.load(_gcp_bootstrap)
    return _modules

def not_implemented_bootstrap(env, _modules=None):
    raise NotImplementedError()

def _common_bootstrap(env, _modules=None):
    if _modules is None:
        _modules = IncrementalModuleLoader()

    _modules.update(_modules=_modules, env=env)
    
    deployment_config = dict(
        _provider=None,
        _force_enable_auth=False,
        version=None,
        stage=None
    )

    stage = env.SPINTOP_STAGE
    db_stage = env.SPINTOP_DATABASE_STAGE

    for _stage in (stage, db_stage):
        if _stage not in ['prod', 'dev', 'test', 'staging']:
            raise ValueError(f'Invalid stage name: {_stage!r}')
    
    deployment_config['stage'] = stage
    deployment_config['db_stage'] = db_stage
    deployment_config['_provider'] = env.SPINTOP_CLOUD_PROVIDER

    _modules.update(
        deployment_config=deployment_config,
        versions=dict(
            spintop=spintop_version
        )
    )
    return _modules

def _gcp_bootstrap(env, deployment_config, _modules=None):
    from .services.param_provider import GCPParamProvider
    from .services.pub_sub import GCPPubSub
    from spintop.services.analytics_provider import BigQueryAnalyticsProvider

    project_id = env.GOOGLE_CLOUD_PROJECT
    pub_sub_emulator = env.PUBSUB_EMULATOR_HOST

    deployment_config['version'] = env.GAE_VERSION
    deployment_config['pubsub_emulator'] = pub_sub_emulator is not None
    deployment_config['_force_enable_auth'] = True

    deployment_config['_gcp'] = {
        'project_id': project_id
    }
    
    _modules.update(project_id=project_id)

    _modules.load(param_provider=GCPParamProvider)
    if deployment_config['stage'] in ['prod', 'staging']:
        # No defaults tolerated when running in prod or staging.
        # Avoids accidental default secret key values for example.
        _modules['param_provider'].tolerate_defaults = False

    
    _modules.load(pub_sub=GCPPubSub)
    _modules.load(tracer_export=initialize_gcp_tracer_export)
    _modules.load(analytics_provider=BigQueryAnalyticsProvider)
    return _modules


import xmltodict
from collections.abc import Mapping, Sequence
import dateutil.parser
import pytz

from spintop.models import (
    SpintopTreeTestRecordBuilder
)
from spintop.transforms import Transformer

from ...logs import _logger

logger = _logger('collectors.niatml')

class NiATMLFormatTransformer(Transformer):
    def __call__(self, atml_filename):
        test_record = collect_filename(atml_filename)
        return self.parse_xml_dict(test_record)

    def parse_xml_dict(self, test_record):
        test_results = test_record['trc:TestResultsCollection']['trc:TestResults']
        dut_id = test_results['tr:UUT']['c:SerialNumber']
        result_set = test_results['tr:ResultSet']
        start_datetime = dateutil.parser.parse(result_set.get('@startDateTime')).replace(tzinfo=pytz.utc)
        end_datetime = dateutil.parser.parse(result_set.get('@endDateTime')).replace(tzinfo=pytz.utc)
        outcome = 'Pass' in result_set.get('tr:Outcome', {}).get('@value')
        
        logger.info('Transforming %s on %s' % (dut_id, start_datetime))
        
        builder = SpintopTreeTestRecordBuilder()
        builder.set_top_level_information(
            start_datetime=start_datetime,
            dut=dut_id,
            testbench='testbench',
            duration=(end_datetime-start_datetime).total_seconds(),
            outcome=outcome
        )
        
        collect_all_tests(collect_phase(result_set), builder)
        
        record = builder.build()
        return record

def collect_all_tests(result_set, builder, depth=0):
    if isinstance(result_set, Mapping):
        with attach_phase(result_set, builder) as builder:
            collect_measures(result_set, builder)
            
            # Collect phases to extract start datetime
            # We sort by datetime because Test and TestGroup order were mixed up on xml parse.
            
            tests = collect_one_or_many_phases(result_set.get('tr:Test', []))
            test_groups = collect_one_or_many_phases(result_set.get('tr:TestGroup', []))
            
            phases = sorted(tests + test_groups, key = lambda phase: phase['start_datetime'])
            
            for test_phase in phases:
                collect_all_tests(test_phase, builder, depth=depth+1)

def collect_one_or_many_phases(array):
    return [collect_phase(test_phase) for test_phase in as_array(array)]
    
def as_array(array):
    if not isinstance(array, Sequence):
        # sometimes contains only one test, which will appear as a dict instead of a sequence.
        array = [array]
    return array
    


def make_measure(result):
    return {
        'value': make_value(result)
    }

def make_params(result):
    params = result.get('tr:Parameters', {})
    if params:
        datum = params.get('tr:Parameter', {}).get('tr:Data', {}).get('c:Datum', {})
        value = datum.get('c:Value').split('\\')[-1]
        return [value]
    else:
        return []
    
def make_value(result):
    TYPES = {
        'ts:TS_double': float,
    }
    datum = result.get('tr:TestData', {}).get('c:Datum', {})
    value = datum.get('@value')
    if value is None:
        value = datum.get('c:Value')
    ts_type = datum.get('@xsi:type')
    return TYPES.get(ts_type, lambda x:x)(value)

def collect_measures(test_phase, builder):
    name = test_phase.get('@name')
    
    # Share the outcome with the surrounding test phase
    # which is usually an AcceptanceCriteria
    outcome = make_outcome(test_phase)
        
    results = as_array(test_phase.get('tr:TestResult', []))
    measures = {}
    
    for result in results:
        measure_name = name + ':' + result.get('@name')
        measures[measure_name] = make_measure(result)

    for measure_name, measure_result in measures.items():
        measure = builder.new_measure(
            name=measure_name,
            outcome=outcome,
            value=measure_result['value']
        )
        

def make_outcome(test_phase):
    
    is_skip = 'Skip' in test_phase.get('tr:Outcome', {}).get('@qualifier', '')
    is_pass = is_skip or 'Pass' in test_phase.get('tr:Outcome', {}).get('@value')
    return dict(
        is_pass=is_pass,
        is_skip=is_skip
    )

def collect_phase(test_phase, accept_all=False):
    name = '#'.join([test_phase.get('@name').split('\\')[-1]] + make_params(test_phase))
    
    return dict(
        name = name,
        start_datetime = dateutil.parser.parse(test_phase.get('@startDateTime')),
        end_datetime = dateutil.parser.parse(test_phase.get('@endDateTime')),
        outcome = make_outcome(test_phase),
        **test_phase
    )
    
def attach_phase(collected_phase, builder):
    return builder.new_phase(
        name=collected_phase['name'],
        outcome=collected_phase['outcome'],
        duration=(collected_phase['end_datetime']-collected_phase['start_datetime']).total_seconds(),
    )

def collect_filename(file_stream):
    data = file_stream.read()
    return xmltodict.parse(data)


import random
import time
import datetime

from spintop.models import SpintopTreeTestRecordBuilder
from spintop.utils import utcnow_aware
from .base import Generator

def normal(mu, sigma):
    return random.normalvariate(mu, sigma)

def random_failure(failure_odd):
    return random.uniform(0, 1.0) < failure_odd

TEST_NAME_GEN = lambda index: 'basicgen_test'
DUT_ID_GEN = lambda index: 'dut%d' % ((index%25) if 17 < (index%25) < 20 else index % 17)
DURATION_GEN = lambda index: 1.0 if (index%25) == 10 or (index%25) == 17 else 0.2


class RandomTestGenerator(Generator):

    def __call__(self, count=25, start_datetime=None, test_phases=('phase1', 'phase2'), measures=(), failure_rate=0.1, **generators):
        test_name_gen = generators.get('test_name_gen', TEST_NAME_GEN)
        dut_id_gen = generators.get('dut_id_gen', DUT_ID_GEN)
        duration_gen = generators.get('duration_gen', DURATION_GEN)
        fail_gen = generators.get('fail_gen', lambda index: random_failure(failure_rate))


        if start_datetime is None:
            start_datetime = utcnow_aware()
            
        for index in range(count):
            builder = SpintopTreeTestRecordBuilder()
            duration = duration_gen(index)
            fails = fail_gen(index)
            builder.set_top_level_information(
                start_datetime=start_datetime,
                dut=dut_id_gen(index),
                testbench=test_name_gen(index),
                duration=duration,
                outcome=dict(
                    is_pass=not fails
                )
            )
            for test_name in test_phases:
                builder.new_phase(test_name, not fails, duration=duration/2)

            for measure_name in measures:
                builder.new_measure(measure_name, True, random.random())

            start_datetime += datetime.timedelta(seconds=duration)
            yield builder.build()



def DeprecatedBy(new_cls):
    class _Temp():
        def __new__(cls, **data):
            data = cls.migrate(data)
            return new_cls(**data)
        
        @classmethod
        def migrate(cls, data):
            raise NotImplementedError()

    _Temp.__name__ = f'TransformTo{new_cls.__name__}'
    return _Temp


from spintop.analytics import AbstractSingerTarget, AnalyticsResponse
from spintop.models import get_json_serializer
from spintop.persistence.base import PersistenceFacade

from .persistence_facade import default_spintop_api

class SpintopAPIAnalytics(AbstractSingerTarget):
    
    def __init__(self, spintop_api=None, uri=None, database_name=None, env=None):
        super().__init__()

        if spintop_api is None:
            spintop_api = default_spintop_api(env, uri, database_name)

        self.spintop_api = spintop_api
        spintop_api.register_analytics(self)
        self.serializer = get_json_serializer()

    @property
    def session(self):
        return self.spintop_api.session

    def send_messages(self, messages_str, request_id=None, stream_name=None):
        return self._stream_named_analytics_endpoint('analytics.stream_update', messages_str, request_id=request_id, stream_name=stream_name)

    def update_named_stream(self, name, raw_data, stream_type=None):
        data = self.serializer.serialize_barrier(raw_data, validate=True)
        endpoint_name = f'analytics.{name}_update'
        return self._stream_named_analytics_endpoint(endpoint_name, data)

    def _stream_named_analytics_endpoint(self, endpoint_name, data, request_id=None, stream_name=None):
        response_data = self.session.put(self.spintop_api.get_link(endpoint_name), json=data).json()
        response = self.serializer.deserialize(response_data)
        return self.complete_response(response, request_id=request_id, stream_name=stream_name)


from spintop.models import BaseDataClass

class AnalyticsResponse(BaseDataClass):
    count: int = 0
    request_id: str = None
    stream_name: str = None

class BigQueryAnalyticsResponse(AnalyticsResponse):
    project_id: str = None
    dataset_id: str = None

    def query(self, deserialize=True):
        # import here to not require bigquery unless imported.
        from .bigquery import query_analytics_response
        return list(query_analytics_response(self, deserialize=deserialize))


from typing import List
from enum import Enum
from datetime import datetime

from spintop.models import (
    BaseDataClass,
    PersistenceIDRecord,
    OutcomeData,
    VersionIDRecord
)

class LCTypes():
    DUT = 'dut'
    STATION = 'station'
    OPERATOR = 'operator'
    SOFTWARE = 'software'
    STEP = 'step'

class LCUnit(VersionIDRecord):
    lc_type: str = ''

    @classmethod
    def from_version_id(cls, version_id, lc_type):
        return cls(
            id=version_id.id,
            version=version_id.version,
            lc_type=lc_type
        )

class LCStep(PersistenceIDRecord):
    outcome: OutcomeData = None
    duration: float = None
    lc_type: str = None # The lifecycle of this step. Linked with the units lc_types.
    units: List[LCUnit] = list
    filename: str = None
    pipeline_uuid: str = None
    
    def __post_init__(self):
        super().__post_init__()
        if self.outcome is not None:
            self.outcome = OutcomeData.create(self.outcome)



from typing import List, Dict, Any
from uuid import uuid4

from spintop.env import SpintopEnv, alias_for
from spintop.messages import SpintopMessagePublisher, Topics
from spintop.models.lifecycles import FileHandlingContext, reserve_lifecycle_streams

from .bootstrap import gcp_bootstrap, no_provider_bootstrap, not_implemented_bootstrap

from .services.dbt_cloud import DbtCloudInterface

class SpintopDeployEnv(SpintopEnv):
    _env_side_input = None

    SPINTOP_DEPLOY_UUID: str = None

    SPINTOP_STAGE: str = 'dev'
    SPINTOP_DATABASE_STAGE: str = 'dev'
    SPINTOP_CLOUD_PROVIDER: str = 'none'

    # GCP
    GOOGLE_CLOUD_PROJECT: str = None
    GAE_VERSION: str = None # Google App Engine
    PUBSUB_EMULATOR_HOST: str = None # Local emulator if set
    COMMIT_SHA: str = None

    # Dbt Cloud
    DBT_CLOUD_AUTH_TOKEN: str = None

    @property
    def uuid(self):
        if self.SPINTOP_DEPLOY_UUID is None:
            self.SPINTOP_DEPLOY_UUID = uuid4()
        return str(self.SPINTOP_DEPLOY_UUID)

    def provider_modules(self):
        return ProviderModules(self.deployment_provider_modules().modules)

    def deployment_provider_modules(self):
        providers = {
            'gcp': gcp_bootstrap,
            'aws': not_implemented_bootstrap,
            'none': no_provider_bootstrap
        }
        cloud_provider = self.SPINTOP_CLOUD_PROVIDER
        bootstrap_fn = providers.get(cloud_provider)

        if bootstrap_fn:
            return bootstrap_fn(self)
        else:
            raise ValueError(f'Invalid cloud provider: SPINTOP_CLOUD_PROVIDER={cloud_provider!r}')

    def file_handling_context(self, filename):
        return FileHandlingContext(filename, self)

    def dbt_cloud_factory(self):
        return DbtCloudInterface(self)

class ProviderModules(object):
    def __init__(self, _modules):
        self._modules = _modules

    def __getattr__(self, name):
        return self._modules[name]

    def get_secret(self, secret_name):
        param_provider = self.param_provider
        return param_provider.get_environ_or_param_value(secret_name)



import dateutil.parser
import pytz

from json import dumps, loads

from copy import copy
from typing import Union, NamedTuple
from collections import OrderedDict, defaultdict
from collections.abc import Sequence, Mapping, MappingView
from functools import lru_cache
from datetime import datetime

from marshmallow import Schema, fields, pre_load, post_dump, EXCLUDE, ValidationError
from marshmallow_jsonschema import JSONSchema

from spintop.utils.marshmallow_dataclass import class_schema
from spintop.utils.marshmallow_oneofschema import OneOfSchema

from spintop.utils import local_tz, isnan, dict_ops

_TYPE_TO_CLS_MAP = {}
_SCHEMA_TYPES = {}

TYPE_KEY = '_type'

_TYPE_CLS_CHILDREN = defaultdict(set)

_REGISTERED_TYPES = []

def register_type(cls, _type=None, schema_cls=None):
    if _type is None:
        _type = cls.__name__
    cls._type = _type
    
    if _type in _TYPE_TO_CLS_MAP and _TYPE_TO_CLS_MAP[_type] != cls:
        raise ValueError("Type %s is already registered to class %s" % (_type, _TYPE_TO_CLS_MAP[_type]))

    _TYPE_TO_CLS_MAP[_type] = cls

    _REGISTERED_TYPES.append((cls, schema_cls))

    for schema_key in _SCHEMA_TYPES:
        Schema = DataClassSchema(cls, schema_type=get_base_schema_cls(schema_key), schema_cls=schema_cls)

        # Building the schema now improves serialization performance by up to 800%
        _SCHEMA_TYPES[schema_key].type_schemas[_type] = Schema()

    for other_cls in _TYPE_CLS_CHILDREN:
        if other_cls in cls.__mro__:
            _TYPE_CLS_CHILDREN[other_cls].add(cls)
    _TYPE_CLS_CHILDREN[cls] = _TYPE_CLS_CHILDREN[cls]

def cls_subclasses(cls):
    """Returns a set containing the subclasses of cls 
    (if cls was at some point passed to register_type as argument)"""
    return _TYPE_CLS_CHILDREN[cls]

def cls_of_obj(obj):
    if isinstance(obj, Mapping):
        return cls_of_serialized(obj)
    else:
        return obj.__class__
    # try:
    #     _type = type_of(obj)
    #     cls = cls_of_type(_type)
    #     if cls is None:
    #         raise AttributeError(f'{cls} is not mapped to a type')
    #     return cls
    # except AttributeError:
    #     return cls_of_serialized(obj)

def cls_of_serialized(obj_serialized):
    return cls_of_type(serialized_type_of(obj_serialized))

def cls_of_type(_type):
    return _TYPE_TO_CLS_MAP.get(_type, None)

def type_of(cls):
    return cls._type

def serialized_type_of(obj_serialized):
    return obj_serialized[TYPE_KEY]

def type_dict_of(cls):
    return {TYPE_KEY: type_of(cls)}

def is_type_of(cls, base_cls):
    return type_of(cls) == type_of(base_cls)

def is_serialized_type_of(obj_serialized, base_cls):
    return serialized_type_of(obj_serialized) == type_of(base_cls)

def get_serializer(schema_key=None):
    return Serializer(get_base_schema_cls(schema_key))

@lru_cache(maxsize=None)
def DataClassSchema(cls, schema_type=None, schema_cls=None):
    if schema_cls:
        schema_cls = compose_schema_cls(schema_type, schema_cls)
    else:
        schema_cls = schema_type
    schema = class_schema(cls, base_schema=schema_cls)
    return schema

def get_base_schema_cls(schema_key):
    try:
        return _SCHEMA_TYPES[schema_key]
    except KeyError:
        raise ValueError(f'Schema key {schema_key!r} not in supported list: {_SCHEMA_TYPES.keys()!r}')

def compose_schema_cls(base, child):
    return type(f'Composed{child.__name__}', (base, child), {})

class Serializer(object):
    def __init__(self, base_schema_cls):
        self.base_schema_cls = base_schema_cls

    def schema(self, *args, cls=None, strict_casting=None, **kwargs):
        schema = self.base_schema_cls(*args, **kwargs)
        
        if strict_casting is not None:
            schema.context['strict_casting'] = strict_casting
        if cls:
            schema.context.update(type_dict_of(cls))
        
        return schema

    def serialize_barrier(self, raw_records, validate=True, cls=None):
        raw_records = list(raw_records)
        schema = self.schema(many=True, cls=cls)

        if raw_records and not isinstance(raw_records[0], Mapping):
            records = schema.dump(raw_records)
        else:
            records = raw_records

        if validate:
            validation = schema.validate(records)
            if validation:
                raise ValidationError(validation)
        
        return records

    def serialize(self, obj, cls=None, strict_casting=None):
        master_schema = self.schema(cls=cls, strict_casting=strict_casting)

        try:
            type_of(obj) # Tests if it has a _type attribute.
        except (AttributeError, TypeError):
            return self.serialize_primitive(obj)
        else:
            return master_schema.dump(obj)

    def deserialize(self, obj, cls=None, strict_casting=None, **kwargs):
        master_schema = self.schema(cls=cls, strict_casting=strict_casting)
        return master_schema.load(obj, **kwargs)

    def serialize_primitive(self, obj):
        if isinstance(obj, Sequence) and not isinstance(obj, str):
            return [self.serialize(subobj) for subobj in obj]
        elif isinstance(obj, Mapping):
            return {key: self.serialize(value) for key, value in obj.items()}
        else:
            try:
                Field = self.base_schema_cls.TYPE_MAPPING[type(obj)]
            except KeyError:
                # No mapped field. Return as-is !
                return obj
            field = Field()
            return field._serialize(obj, None, None)

    def deserialize_primitive(self, obj, cls=None):
        if isinstance(obj, Sequence):
            return [self.deserialize(subobj, cls=cls) for subobj in obj]
        elif isinstance(obj, Mapping):
            return {key: self.deserialize(value, cls=cls) for key, value in obj.items()}
        else:
            raise ValueError('Unable to deserialize primitives other than list or dict.')


class NoopDatetime(fields.AwareDateTime):
    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        return value

class OrderedDictField(fields.Mapping):
    def _serialize(self, value, attr, obj, **kwargs):
        value_as_dict = dict(value)
        serialized_as_dict = super()._serialize(value_as_dict, attr, obj, **kwargs)
        return [(key, serialized_as_dict[key]) for key in value]

    def _deserialize(self, list_value, attr, data, **kwargs):
        value_as_dict = {key: value for key, value in list_value}
        deserialized_as_dict = super()._deserialize(value_as_dict, attr, data, **kwargs)
        result = OrderedDict()
        for key, _ in list_value:
            result[key] = deserialized_as_dict[key]
        return result

class FloatFieldNoNaN(fields.Float):
    def _serialize(self, value, attr, data, **kwargs):
        if value is None or isnan(value):
            value = None
        num = super()._serialize(value, attr, data, **kwargs)
        return num

class AwareDateTimeDeserializeNoop(fields.AwareDateTime):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, default_timezone=pytz.utc, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime):
            return value.replace(tzinfo=self.default_timezone)
        else:
            return super()._deserialize(value, attr, data, **kwargs)

class DictJsonField(fields.Dict):
    def _serialize(self, value, attr, obj, **kwargs):
        if value:
            return dumps(super()._serialize(value, attr, obj, **kwargs))
        else:
            return '{}' # Empty json

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return {}
        else:
            return super()._deserialize(loads(value), attr, data, **kwargs)

def base_type_mapping(additions={}):
    default_type_mapping = copy(Schema.TYPE_MAPPING)
    default_type_mapping.update({
        float: FloatFieldNoNaN,
        OrderedDict: OrderedDictField,
        datetime: AwareDateTimeDeserializeNoop,
        list: fields.List,
        set: fields.List,
        tuple: fields.Tuple,
        dict: fields.Dict
    })

    try:
        from pandas import Timestamp
    except ImportError:
        pass
    else:
        default_type_mapping.update({
            Timestamp: AwareDateTimeDeserializeNoop
        })

    default_type_mapping.update(additions)
    return default_type_mapping

class AbstractSchema(OneOfSchema):
    class Meta:
        unknown = EXCLUDE
        ordered = True

    type_schemas = {}
    TYPE_MAPPING = base_type_mapping()
    type_field = TYPE_KEY
    _attached = False

    _type = fields.String(dump_only=True)

    def __new__(cls, *args, **kwargs):
        if not cls._attached:
            cls.attach_registered_dataclasses()
        obj = super().__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj

    def get_obj_type(self, obj):
        # If a mapping, returns the class from the _type attribute
        # If an object, returns the class itself.
        cls = cls_of_obj(obj) 
        if cls is None:
            return None
        else:
            # Return the string registered to that cls.
            return type_of(cls)

    @classmethod
    def build_schema_type(cls, name_or_base_cls, additional_type_mapping=None, _type_mapping=None):
        if _type_mapping is None:
            if additional_type_mapping is None:
                additional_type_mapping = {}
            _type_mapping = base_type_mapping(additional_type_mapping)

        cls_dict = {
            'type_schemas': {},
            'TYPE_MAPPING': _type_mapping
        }

        if isinstance(name_or_base_cls, type):
            name = name_or_base_cls.__name__
            base_cls = (cls, name_or_base_cls)
        else:
            name = name_or_base_cls
            base_cls = (cls,)

        schema_type = type(name, base_cls, cls_dict)
        schema_type.attach_registered_dataclasses()
        return schema_type
    
    @classmethod
    def attach_registered_dataclasses(cls):
        cls.type_schemas = {}
        cls._attached = True

        for data_cls, schema_cls in _REGISTERED_TYPES:
            Schema = DataClassSchema(data_cls, schema_type=cls, schema_cls=schema_cls)

            # Building the schema now improves serialization performance by up to 800%
            cls.type_schemas[data_cls._type] = Schema()
        


def register_schema_type(key, schema_type, _default=False):
    if key in _SCHEMA_TYPES:
        raise ValueError(f'Key {key!r} is already registered.')

    _SCHEMA_TYPES[key] = schema_type

    if _default:
        _SCHEMA_TYPES[None] = schema_type

    return schema_type
    

JsonCompatSchema = AbstractSchema.build_schema_type(
    'JsonCompatSchema'
)

BsonCompatSchema = AbstractSchema.build_schema_type(
    'BsonCompatSchema',
    {
        datetime: NoopDatetime # keeps datetime as is
    }
)

DeepTabularCompatSchema = AbstractSchema.build_schema_type(
    'DeepTabularCompatSchema',
    {
        dict: DictJsonField # Serialize dicts to JSON. Required to have a constant schema.
    }
)

register_schema_type('tabular', DeepTabularCompatSchema)
register_schema_type('json', JsonCompatSchema)
register_schema_type('bson', BsonCompatSchema, _default=True)



def get_bson_serializer():
    return get_serializer('bson')

def get_json_serializer():
    return get_serializer('json')

class SerializedWrapper(dict_ops.DictWrapper):
    
    @property
    def type_(self):
        return self[TYPE_KEY]

    @type_.setter
    def type_(self, value):
        self[TYPE_KEY] = value

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

## JSON Schema generation

@lru_cache(maxsize=None)
def dump_json_schema(cls):
    schema = cls.get_schema()()
    json_schema_generator = JSONSchema(props_ordered=True)
    return expand_schema_refs(json_schema_generator.dump(schema))

def expand_schema_refs(json_schema, _original=None):
    if _original is None:
        _original = json_schema
    
    if '$ref' in json_schema:
        definition_path = json_schema['$ref'].split('/')[1:] # Remove first '#'
        reference_content = _original
        for p in definition_path:
            reference_content = reference_content[p]

        json_schema.update(reference_content)
        json_schema.pop('$ref')
    
    json_schema.pop('$schema', None)

    for key, value in json_schema.items():
        if isinstance(value, Mapping):
            json_schema[key] = expand_schema_refs(value, _original=_original)

    return json_schema




import prefect
import time

from concurrent.futures import ThreadPoolExecutor

from prefect.cli.execute import _execute_flow_run
from prefect.utilities.graphql import GraphQLResult
from prefect.environments.storage import GCS, S3, Azure, Local, GitHub, GitLab, Webhook
from prefect.serialization.storage import StorageSchema

class ServerlessAgent(prefect.agent.local.LocalAgent):
    def __init__(self, *args, **kwargs):
        kwargs['max_polls'] = kwargs.get('max_polls', 3)
        kwargs['hostname_label'] = False
        super().__init__(*args, **kwargs)

    def run_flow_id(self, flow_id, **kwargs):
        self.client.create_flow_run(version_group_id=flow_id, **kwargs)

        self.client.attach_headers({"X-PREFECT-AGENT-ID": self._register_agent()})

        with ThreadPoolExecutor(max_workers=1) as executor:
            flow_executed = self.agent_process(executor)
            if not flow_executed:
                time.sleep(2)
                flow_executed = self.agent_process(executor)
            
            if not flow_executed:
                raise RuntimeError('Flow did not execute.')


    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy flow runs on your local machine as Docker containers

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment

        Raises:
            - ValueError: if deployment attempted on unsupported Storage type
        """
        self.logger.info("Deploying flow run {}".format(flow_run.id))  # type: ignore

        if not isinstance(
            StorageSchema().load(flow_run.flow.storage),
            (Local, Azure, GCS, S3, GitHub, GitLab, Webhook),
        ):
            self.logger.error(
                "Storage for flow run {} is not a supported type.".format(flow_run.id)
            )
            raise ValueError("Unsupported Storage type")

        prefect.context.flow_run_id = flow_run.id
        prefect.context.flow_id = flow_run.flow.id
        prefect.config.logging.log_to_cloud = str(self.log_to_cloud).lower()
        prefect.config.engine.flow_runner.default_class = "prefect.engine.cloud.CloudFlowRunner"
        prefect.config.engine.task_runner.default_class = "prefect.engine.cloud.CloudTaskRunner"

        _execute_flow_run()

        return "OK"



from spintop.analytics.bigquery import AdvancedBigQueryAnalytics

class BigQueryAnalyticsProvider(object):
    def __init__(self, env, project_id):
        self.project_id = project_id
        self.env = env

    def analytics_target_factory(self, database_name, namespace=None):
        if namespace is None:
            namespace = self.project_id
        return AdvancedBigQueryAnalytics(uri=namespace, database_name=database_name, env=self.env)



import atexit
import os
import warnings
import dotenv
from datetime import datetime

import pytz
import pytest
import pytest_dependency # pylint: disable=unused-import

from .integration_config import IntegrationConfig

from spintop import Spintop
from spintop.env import SpintopEnv

from spintop.auth import AuthModule, InMemoryCredentialsStore
from spintop.api_client import SpintopAPIClientModule, SpintopAPISpecAuthBootstrap

from spintop.generators.random_gen import RandomTestGenerator
from spintop.models import SpintopTestRecordCollection

dotenv.load_dotenv()

def _integration_file(config):
    return os.path.join(config.rootdir, 'integration_secrets.yml')

@pytest.fixture(scope="session")
def random_generator():
    return RandomTestGenerator()

@pytest.fixture(scope="session")
def auth_module(integration_config):
    return AuthModule(
        auth_bootstrap=SpintopAPISpecAuthBootstrap(integration_config.spintop_api_url) if integration_config else None,
        credentials_store=InMemoryCredentialsStore()
    )

@pytest.fixture(scope="session")
def spintop_env(integration_config):
    if integration_config:
        assert 'SPINTOP_M2M_CLIENT_ID' in os.environ
        assert 'SPINTOP_M2M_SECRET_KEY' in os.environ
        return SpintopEnv( 
            api_url=integration_config.spintop_api_url, 
            org_id=integration_config.expected_org
        )
    else:
        return SpintopEnv()

@pytest.fixture(scope="session", autouse=True)
def spintop_api_module(integration_config, spintop_env):
    if not integration_config:
        return None
    else:
        return spintop_env.spintop_factory()

@pytest.fixture(scope="session", autouse=True)
def spintop_user_api_module(integration_config, auth_module, spintop_env):
    if not integration_config:
        return None
    else:
        return spintop_env.spintop_factory(auth_factory=lambda: auth_module)

@pytest.fixture(scope="module")
def raw_collection(random_generator):
    raw_collection = SpintopTestRecordCollection(
        random_generator.generate(
            count=2, 
            start_datetime=datetime(year=2020, month=1, day=1, tzinfo=pytz.utc), 
            test_phases=('phase1', 'phase2'),
            measures=('measure1', 'measure2'),
            failure_rate=1.0
        )
    )
    return raw_collection

@pytest.fixture(scope="module")
def flat_record_collection(raw_collection):
    return raw_collection

@pytest.fixture()
def flat_test_record(flat_record_collection):
    return flat_record_collection.records[0]

KNOWN_HOSTS = {
    'dev': 'https://tack-infra-dev.nn.r.appspot.com',
    'prod': 'https://cloud.spintop.io',
    None: 'http://127.0.0.1:5079'
}

def pytest_addoption(parser):
    parser.addoption(
        "--test-against", action="store", default=None, help=f"The spintop API url to test against, or one of {set(KNOWN_HOSTS.keys())}"
    )

def pytest_configure(config):
    config.addinivalue_line("markers", "integration: mark test as auth/api integration test")

def pytest_collection_modifyitems(config, items):
    integration_file = _integration_file(config)
    if os.path.exists(integration_file):
        return
    else:
        warnings.warn(f'Integration file {integration_file!r} does not exist; integration with API will not be tested. This file should be located in the pytest rootdir.')
        skip_slow = pytest.mark.skip(reason="need --integration-file option to run")
        for item in items:
            if "integration" in item.keywords:
                item.add_marker(skip_slow)

def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item

def pytest_runtest_setup(item):
    previousfailed = getattr(item.parent, "_previousfailed", None)
    if previousfailed is not None:
        pytest.xfail("previous test failed (%s)" % previousfailed.name)

@pytest.fixture(scope="session")
def integration_config(request):
    filepath = _integration_file(request.config)
    if os.path.exists(filepath):
        config = IntegrationConfig.load_file(filepath)
        host = request.config.getoption("--test-against")
        host = KNOWN_HOSTS.get(host, host)
        config.spintop_api_url = host
        create_atexit_print_test_against(host)
    else:
        config = None
    return config

def create_atexit_print_test_against(spintop_api_url):
    def report():
        print(f"Tested against {spintop_api_url!r}")
    atexit.register(report)

@pytest.fixture()
def spintop():
    return Spintop()

@pytest.fixture()
def flat_test_record(flat_record_collection):
    return flat_record_collection.records[0]


import logging

from collections.abc import Mapping
from contextlib import contextmanager

from prefect import task

from .messages import Message, encode_data

logger = logging.getLogger(__name__)

@contextmanager
def cloud_functions_context(event, context):
    try:
        if context:
            logger.info(
                'This Function was triggered by '
                f'message ID {context.event_id} of type {context.event_type} '
                f'published at {context.timestamp}'
            )
        else:
            logger.info('No context information surrounding function trigger.')
        yield
    except: # pylint: disable=bare-except
        logger.exception('Exception occured', exc_info=True)
        try_report_active_exc()

def try_report_active_exc():
    try:
        from google.cloud import error_reporting
        client = error_reporting.Client()
        client.report_exception()
    except: # pylint: disable=bare-except
        logger.warning('Unable to report to google error reporting', exc_info=True)

class PubSubRouter():
    def __init__(self, handlers=None):
        if handlers is None:
            handlers = []
        self.handlers = handlers

    def cloud_functions_handler(self):
        return self._handle_cloud_functions

    def _handle_cloud_functions(self, event, context):
        with cloud_functions_context(event, context):
            print(f'event={event}')
            message = Message(event=event, context=context)
            self.handle(message)

    def pub_sub_handler(self):
        return self._handle_pub_sub

    def _handle_pub_sub(self, pub_sub_message):
        message = Message(event={'data': pub_sub_message.data}, context=None)
        self.handle(message)

    def handle(self, message):
        handled = False
        for handler in self.handlers:
            handler_message = handler.filter(message)
            if handler_message:
                print(f'Using handler {handler!r}')
                handled = True
                break_handling = handler.handle(handler_message)
                if break_handling:
                    break
        
        if not handled:
            print(f'No handlers for message')


class PubSubTrigger(object):
    def __init__(self, topic):
        self.topic = topic
        from google.cloud import pubsub
        self._client = pubsub.PublisherClient() 

    @property
    def client(self):
        return self._client

    def trigger(self, message):
        print(f'Triggering {message!r}')
        return self.trigger_event(message.event)

    def trigger_event(self, event):
        event = encode_event(event)
        data_encoded = event['data']
        attributes = event['attributes']
        print(f'Publishing data={data_encoded!r}, attributes={attributes!r}')
        self.client.publish(self.topic, data_encoded, **attributes)

def encode_event(event):
    data = event.get('data', {})
    attributes = event.get('attributes', {})
    return dict(
        data=encode_data(data),
        attributes=attributes
    )
    



from marshmallow import Schema, fields

class CredentialsSchema(Schema):
    username = fields.String(allow_none=True)
    access_token = fields.String()
    refresh_token = fields.String(allow_none=True)
    org_id = fields.String(allow_none=True)
    refresh_module = fields.String()
    client_id = fields.String(allow_none=True)
    
credentials_schema = CredentialsSchema(partial=True)